"""Abstract base class for tide data.

Copyright (C) 2023  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from abc import ABC, abstractmethod
import datetime as dt

import pandas as pd


class AbstractTideData(ABC):
    """Abstract base class for tidal data."""

    def get_tidal_gravity_correction(self, epoch_dt):
        pass

    @property
    @abstractmethod
    def get_tide_corr_df(self):
        """Returns the complete tidal gravity time-series as pandas DataFrame"""
        pass

    @property
    @abstractmethod
    def filename(self) -> str:
        """Returns the name of the data file."""
        pass

    @property
    @abstractmethod
    def filetype(self) -> str:
        """Returns the type of the data file."""
        pass

    @property
    @abstractmethod
    def number_channels(self) -> str:
        """Returns the number of data channels."""
        pass

    @property
    @abstractmethod
    def number_records(self) -> str:
        """Returns the number of data records."""
        pass

    @property
    @abstractmethod
    def starttime(self) -> dt.datetime:
        """Returns the start time of the tidal timeseries."""
        pass

    @property
    def starttime_str(self) -> str:
        """Returns the start time of the tidal timeseries as string."""
        return self.starttime.strftime('%Y-%d-%d %H:%M:%S')

    @property
    @abstractmethod
    def endtime(self) -> dt.datetime:
        """Returns the end time of the tidal timeseries."""
        pass

    @abstractmethod
    def get_channel_df(self, channel: [int, str]) -> pd.DataFrame:
        """Return the specified channel and the time reference (DateTime) as pandas DataFrame.

        Parameters
        ----------
        channel : str or int or 'last'
            If `channel` is as string the argument is interpreted as channel name and the according channel data is
            returned. If `channel` is an integer, it is interpreted as the number of the channel to be returned. `last`
            indicates that the last channel in the TSF file should be returned.


        Returns
        -------
        pandas.DataFrame containing the data records of the selected channel and the time reference as DateTime.
        """
        pass

    @property
    def endtime_str(self) -> str:
        """Returns the end time of the tidal timeseries as string."""
        return self.endtime.strftime('%Y-%d-%d %H:%M:%S')

    @property
    def filename_without_path(self) -> str:
        """Returns the filename without path."""
        pass

    def __str__(self):
        return f'Tidal data loaded from {self.filename} ({self.filetype}) with {self.number_channels} channels ' \
               f'and {self.number_records} datasets ({self.starttime_str} to {self.endtime_str})'







"""Class for correction data provided as time series per station and survey.

Copyright (C) 2023  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass
import typing
from datetime import datetime, timezone
import warnings

import pandas as pd
import scipy
import numpy as np

import gravtools.tides.correction_time_series
from gravtools.tides.tide_data_tfs import TSF
from gravtools import settings

class CorrectionTimeSeries:
    """Class for correction data provided as time series per station and survey.

    Attributes
    ----------
    surveys: dict of `gravtools.tides.correction_time_series.SurveyCorrections`
        Dict of `SurveyCorrection` objects with ths survey names as keys.
    """

    def __init__(self):
        """Default initializer."""
        self.surveys = {}

    def load_tfs_file(self, survey_name: str, filename_tsf: str, location: str ='', instrument: str ='',
                      data_type: str ='', overwrite_channel: bool=True, is_correction: bool='False'):
        """Load time series data from a TSoft TSF file for one survey.

        Parameters
        ----------
        survey_name: str
            Name of the survey for which the data is loaded.
        filename_tsf: str
            Name and path of the TSF file from which the correction time series data is loaded.
        location: str, optional (default='')
            If not empty, only channels with matching locations are loaded.
        instrument: str, optional (default='')
            If not empty, only channels with matching instruments are loaded.
        data_type: str, optional (default='')
            If not empty, only channels with matching data types are loaded.
        overwrite_channel: bool, optional (default=`True`)
            `True` implies that an existing StationCorrection object of a station will be overwritten.
        is_correction: bool, optional (default=`False`)
            `True` implies that the channel data loaded from the TSF file model the gravity effect of phenomena, rather
            than corrections for gravity observations. Both options have the opposite sign: while corrections have to be
            added to gravity observations, effects have to be subtracted, in order to reduce observations.
            See TSoft manual (version 2.2.4 Release date 2015-09-09), p. 15: "The loading calculation by Tsoft will
            result in the correction, not the effect.
            On the other hand the prediction of the solid Earth tides using the WDD parameter set
            provided by Tsoft directly will result in the effect. Therefore both must be treated with
            different sign in order to reduce a gravity time series correctly."
        """
        # Load TSF file:
        tsf_data = TSF.from_tfs_file(filename_tsf)

        # Loop over channel and retrieve required data:
        retrieved_channel_locations = []
        retrieved_channel_numbers = []
        stations_corrections_dict = {}
        for channel_metadata in tsf_data.channel_metadata:
            ch_channel_name = channel_metadata.channel_name
            ch_instrument = channel_metadata.instrument_name
            ch_data_type = channel_metadata.data_type
            ch_location = channel_metadata.location
            channel_number = channel_metadata.channel_number

            if location:
                if location != ch_location:
                    continue
            if instrument:
                if instrument != ch_instrument:
                    continue
            if data_type:
                if data_type != ch_data_type:
                    continue
            if ch_location in retrieved_channel_locations:
                raise RuntimeError(f'Ambiguous TSF channel matching! The channels '
                                   f'{retrieved_channel_numbers[retrieved_channel_locations.index(ch_location)]} and '
                                   f'{channel_number} share the same location ({ch_location}). '
                                   f'Please constrain the channel selection by providing additional arguments ('
                                   f'instrument, data type) or edit the TSF file.')
            retrieved_channel_locations.append(ch_location)
            retrieved_channel_numbers.append(channel_number)
            ch_epoch_dt, ch_data  = tsf_data.get_channel_np(channel=channel_number)
            time_series = TimeSeries(ref_time_dt=ch_epoch_dt, data=ch_data, unit=channel_metadata.unit,
                                     data_source=f'{tsf_data.filename_without_path} ({tsf_data.filetype})',
                                     description=ch_channel_name,
                                     is_correction=is_correction)


            stations_corrections_dict[ch_location] = StationCorrections(station_name=ch_location,
                                                                        tidal_correction=time_series)

        # Checks:
        if len(retrieved_channel_locations) == 0:
            tmp_str_list = []
            if location:
                tmp_str_list.append(f'station "{location}"')
            if instrument:
                tmp_str_list.append(f'instrument "{instrument}"')
            if data_type:
                tmp_str_list.append(f'data_type "{data_type}"')
            if tmp_str_list:
                raise RuntimeError(f'Could not retrieve any data from TSF file {filename_tsf} with the following '
                                   f'restrictions:' + ', '.join(tmp_str_list) + '.')
            else:
                raise RuntimeError(f'Could not retrieve any data from TSF file {filename_tsf}.')

        # Add loaded data:
        if survey_name in self.surveys.keys():
            # Add station corrections to existing survey correction object:
            for stat_name, stat_corr in stations_corrections_dict.items():
                _ = self.surveys[survey_name].add_station_correction(station_name=stat_name,
                                                                     station_correction=stat_corr,
                                                                     overwrite=overwrite_channel)
        else:
            # Create survey correction object and add the station correction data:
            self.surveys[survey_name] = SurveyCorrections(survey_name=survey_name, stations=stations_corrections_dict)

    def delete_survey_correction(self, survey_name):
        """Removes a survey correction object.

        Parameters
        ----------
        survey_name: str
            Survey name.
        """
        del self.surveys[survey_name]

@dataclass
class TimeSeries:
    """Time series data.

    Parameters
    ----------
    ref_time_dt: `numpy.array` of `datetime` objects
        Reference times for the time series.
    data: `numpy.array`
        Data series. Has to have the same length as `ref_times_dt`
    unit: str
        Unit of the data.
    data_source: str
        Description of the data source, e.g. file name and/or file type, etc.
    description: str, optional (default='')
        Optional description of the time series.
    is_correction: bool, optional (default=`False`)
        `True` implies that the channel data loaded from the TSF file model the gravity effect of phenomena, rather
        than corrections for gravity observations. Both options have the opposite sign: while corrections have to be
        added to gravity observations, effects have to be subtracted, in order to reduce observations.
    created_utc_dt: `datetime` object
        Date and time in UTC of creating this time series, e.g. loading it from a file.

    """
    ref_time_dt: np.array  # np.array of DateTime objects
    data: np.array
    unit: str
    data_source: str
    description: str = ''
    is_correction: bool = True
    # created_utc_dt: datetime = None  # Is initialized in __post_init__

    def __post_init__(self):
        """Executed at the very end of __init__"""
        self.created_utc_dt = datetime.now(tz=timezone.utc)

    @property
    def created_datetime_utc(self):
        """Returns the time and date (in UTC) when the times series was created as datetime object."""
        return self.created_utc_dt

    @property
    def created_datetime_utc_str(self):
        """Returns the time and date (in UTC) when the times series was created as string."""
        return self.created_utc_dt.strftime('%Y-%m-%d %H:%M:%S.%f')

    @property
    def start_datetime(self):
        """Returns the start date and time."""
        return min(self.ref_time_dt)

    @property
    def start_datetime_str(self):
        """Returns the start date and time as string."""
        return pd.to_datetime(self.start_datetime).strftime('%Y-%m-%d %H:%M:%S.%f')

    @property
    def end_datetime(self):
        """Returns the end date and time."""
        return max(self.ref_time_dt)

    @property
    def end_datetime_str(self):
        """Returns the end date and time as string."""
        return pd.to_datetime(self.end_datetime).strftime('%Y-%m-%d %H:%M:%S.%f')

    @property
    def duration_timedelta64_ns(self):
        """Returns the duration as `numpy.timedelta64[ns]` object."""
        return max(self.ref_time_dt) - min(self.ref_time_dt)

    def duration_dhms(self):
        """Returns the duration as days, hours, minutes and seconds."""
        dt = max(self.ref_time_dt) - min(self.ref_time_dt)
        days = dt.astype('timedelta64[D]').astype(int)
        hours = dt.astype('timedelta64[h]').astype(int) - days * 24
        minutes = dt.astype('timedelta64[m]').astype(int) - days*24*60 - hours*60
        seconds = dt.astype('timedelta64[s]').astype(float) - days*24*60*60 - hours*60*60 - minutes*60
        return days, hours, minutes, seconds

    @property
    def duration_dhms_str(self):
        """Returns the duration as string (days, hours, minutes and seconds)"""
        days, hours, minutes, seconds = self.duration_dhms()
        return f'{days} days, {hours} hours, {minutes} minutes, {seconds} seconds'

    @property
    def number_of_datapoints(self):
        """Returns the number of datapoints."""
        return len(self.data)

    @property
    def model_type(self):
        """Returns the model type, i.e. whether effects or corrections are modeled."""
        if self.is_correction:
            return 'Correction'
        else:
            return 'Effect'

    def interpolate(self, interp_times: [np.array, typing.List[datetime]], kind: str = 'quadratic',
                    return_correction: bool = False) -> np.ndarray :
        """Return an interpolated value for the given interpolation epoch.

        Parameters
        ----------
        interp_times: list(datetime) or np.array(datetime)
            Interpolation epoch given as `datetime` object w.r.t. UTC! The interpolation times have to bin within the
            time range of the data series.
        kind: str or int, optional (default='quadratic')
            Specifies, which interpolation method is used. This argument is directly passed to
            `scipy.interpolate.inter1`. For options see scipy reference.
        return_correction: bool, optional (default=`False`)
            If `True`, corrections are returned considering the `self.is_correction` attribute.

        Returns
        -------
        `numpy.ndarray`: Interpolated value for the given interpolation times.
        """
        # Convert datetimes to UNIX timestamps, because numerical values are needed for interpolation:
        interp_times_unix = np.fromiter((t.replace(tzinfo=timezone.utc).timestamp() for t in interp_times), float)
        x = self.ref_time_unix
        y = self.data
        interp_func = scipy.interpolate.interp1d(x, y, kind=kind)
        interp_values = interp_func(interp_times_unix)
        if return_correction:
            if not self.is_correction:
                interp_values = interp_values * -1  # Convert from effect to correction!
        return interp_values

    def to_df(self):
        """Returns the time series as pandas DataFrame with the time reference as index (sorted!)."""
        df = pd.DataFrame({'epoch_dt': self.ref_time_dt,'data': self.data})
        df.set_index('epoch_dt', inplace=True)
        return df

    @property
    def ref_time_unix(self):
        """Returns the reference times as UNIX timestamps (seconds since Jan 1, 1970)."""
        return self.ref_time_dt.astype('int64')/1e9


@dataclass
class StationCorrections:
    """Correction time series data for one station.

    Notes
    -----
    For each station and correction type (e.g. tidal correction) only ONE time series can be added!
    """
    station_name: str
    tidal_correction: TimeSeries


@dataclass
class SurveyCorrections:
    """Correction time series data for one survey with multiple stations."""
    survey_name: str
    stations: typing.Dict[str, StationCorrections]

    def add_station_correction(self, station_name: str, station_correction: StationCorrections, overwrite: bool=True,
                               warn: bool=True):
        """Add a StationCorrection object to `self.stations`.

        Parameters
        ----------
        station_name: str
            Station name.
        station_correction: `StationCorrections`
            StationCorrections object containing correction time series data for a specific station.
        overwrite: bool, optional (default=`True`)
            `True` implies that an existing StationCorrection object of a station will be overwritten.
        warn: bool, optional (default=`True`)
            `True` implies that a warning is issued if an existing StationCorrection will or would bne be overwritten.

        Returns
        -------
        bool: `True`, if the station correction data was added. Otherwise, `False`.
        """
        # Check, if data for this station already exists:
        if station_name in self.stations.keys():
            if not overwrite:
                if warn:
                    warnings.warn(f'Station correction data for station {station_name} already exists! Overwriting '
                                  f'data is not permitted!')
                return False
            else:
                if warn:
                    warnings.warn(f'Station correction data for station {station_name} already exists! Overwriting '
                                  f'data is permitted and the existing data will be overwritten!')
                self.stations[station_name] = station_correction
                return True
        else:
            self.stations[station_name] = station_correction
            return True

    def delete_station_correction(self, station_name):
        """Removes a station correction object.

        Parameters
        ----------
        station_name: str
            Station name.
        """
        del self.stations[station_name]

    @property
    def number_of_stations(self):
        """Returns the number of station correction objects."""
        return len(self.stations)

    @property
    def station_names(self):
        """Returns the names of all stations."""
        return list(self.stations.keys())

@staticmethod
def convert_to_mugal(data, data_unit: str):
    """Converts data to from a given unit to µGal."""
    if data_unit not in settings.UNIT_CONVERSION_TO_MUGAL:
        raise RuntimeError(f'Conversion factor from {data_unit} to µGal not defined. Please add the factor to'
                           f'gravtools.settings.UNIT_CONVERSION_TO_MUGAL.')
    return data * settings.UNIT_CONVERSION_TO_MUGAL[data_unit]











"""Dialog for the calculation of the instrumental drift.

Copyright (C) 2024  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pytz
import datetime as dt
import os


from PyQt5.QtWidgets import QDialog, QMessageBox
from PyQt5 import QtCore
import pyqtgraph as pg
import numpy as np
import pandas as pd

from gravtools.gui.dialog_calc_drift import Ui_Dialog_calculate_drift
from gravtools import settings
from gravtools.gui.gui_misc import get_station_color_dict
from gravtools import __version__

class TimeAxisItem(pg.AxisItem):
    """"Needed to handle the x-axes tags representing date and time.
    From: https://stackoverflow.com/questions/49046931/how-can-i-use-dateaxisitem-of-pyqtgraph

    Notes
    -----
    The timestamps need to refer to UTC!"
    """
    def tickStrings(self, values, scale, spacing) -> str:
        """Handles the x-axes tags representing date and time."""
        return [dt.datetime.fromtimestamp(value, tz=pytz.utc).strftime(settings.Y_TICK_DATETIME_FORMAT) for value in values]

class DialogCalcDrift(QDialog, Ui_Dialog_calculate_drift):
    """Dialog for drift calibration tests."""

    def __init__(self, parent=None):
        super().__init__(parent)
        # Run the .setupUi() method to show the GUI
        self.setupUi(self)

        # Other initial settings:
        self.set_up_drift_plot_widget()

        # Connect signals and slots
        self.comboBox_survey.currentIndexChanged.connect(self.update_station_combobox)
        self.comboBox_station.currentIndexChanged.connect(self.calc_and_plot)
        self.spinBox_degree.valueChanged.connect(self.calc_and_plot)
        self.checkBox_multiple_setups_only.stateChanged.connect(self.update_station_combobox)
        self.checkBox_active_obs_only.stateChanged.connect(self.calc_and_plot)
        self.spinBox_min_no_obs.valueChanged.connect(self.calc_and_plot)
        self.checkBox_remove_g_offset.stateChanged.connect(self.calc_and_plot)

        # Set up combo box for selecting the drift calculation method:
        for idx, (method_name, tooltip) in enumerate(settings.DRIFT_CALC_METHODS.items()):
            self.comboBox_method.addItem(method_name)
            self.comboBox_method.setItemData(idx, tooltip, QtCore.Qt.ToolTipRole)
            if method_name == settings.DRIFT_CALC_DEFAULT_METHOD:
                self.comboBox_method.setCurrentIndex(idx)

        # Init. attributes:
        self.station_color_dict = {}
        self.drift_poly_coeff = {}
        self.drift_start_time = None

    def invoke(self):
        """Invoke dialog."""
        # Get current campaign data:
        self.reset_gui()
        if self.parent().campaign is None:
            pass
        else:
            # Update GUI items based on current campaign data
            self.update_survey_and_station_combobox()
        return self.exec()

    def update_survey_and_station_combobox(self):
        """Update GUI items based on the current campaign data."""
        # Invoke only when executing the dialog.
        # - Silence the related signals meanwhile.
        if self.parent().campaign is None:
            self.reset_gui()
            return
        survey_names = self.parent().campaign.survey_names
        if not survey_names:  # No surveys in campaign
            self.reset_gui()
            return
        selected_survey = self.comboBox_survey.currentText()
        self.comboBox_survey.blockSignals(True)
        for survey_name in survey_names:
            self.comboBox_survey.addItem(survey_name)
        # Select previously selected item, if available!
        if selected_survey:  # Not empty
            idx = self.comboBox_survey.findText(selected_survey)
            if idx != -1:  # Item found!
                self.comboBox_survey.setCurrentIndex(idx)
        self.comboBox_survey.blockSignals(False)
        self.update_station_combobox()

    def update_station_combobox(self):
        """Update the station selection combo box based on the currently selected survey."""
        current_survey_name = self.comboBox_survey.currentText()
        # if not current_survey_name:
        #     return
        current_survey = self.parent().campaign.surveys[current_survey_name]

        station_names = current_survey.observed_stations
        self.station_color_dict = get_station_color_dict(station_names, randomize=True)
        if self.checkBox_multiple_setups_only.isChecked():
            station_names_filtered = []
            for station in station_names:
                tmp_filter = current_survey.obs_df['station_name'] == station
                if self.checkBox_active_obs_only.isChecked():
                    tmp_filter = tmp_filter & current_survey.obs_df['keep_obs']
                if len(current_survey.obs_df.loc[tmp_filter, 'setup_id'].unique()) > 1:
                    station_names_filtered.append(station)
            station_names = station_names_filtered
        station_names = ['All'] + station_names

        selected_station = self.comboBox_station.currentText()
        self.comboBox_station.blockSignals(True)
        self.comboBox_station.clear()
        for station_name in station_names:
            self.comboBox_station.addItem(station_name)
        # Select previously selected item, if available!
        if selected_station:  # Not empty
            idx = self.comboBox_station.findText(selected_station)
            if idx != -1:  # Item found!
                self.comboBox_station.setCurrentIndex(idx)
            else:
                self.comboBox_station.setCurrentIndex(0)  # First item = 'All'
        self.comboBox_station.blockSignals(False)
        self.comboBox_station.currentIndexChanged.emit(self.comboBox_station.currentIndex())

    def reset_gui(self):
        """Reset the GUI dialog and empty all items and plots."""
        # Gui items:
        self.comboBox_survey.blockSignals(True)
        self.comboBox_survey.clear()
        self.comboBox_survey.blockSignals(False)
        self.comboBox_station.blockSignals(True)
        self.comboBox_station.clear()
        self.comboBox_station.blockSignals(False)
        # Plot:
        self.drift_calib_plot.clear()
        self.drift_calib_plot.legend.clear()
        self.drift_calib_plot.setTitle('')
        # Attributes:
        self.drift_poly_coeff = {}
        self.drift_start_time = None

    def calc_and_plot(self):
        """Calculate drift parameters and plot the current data."""
        # Invoke via signal whenever a relevant selection in the GUI dialog changed (send signal!)
        # - Make sure that this method is not executed several times in a row... block signals if required!

        # Reset:
        self.drift_calib_plot.clear()
        self.drift_calib_plot.legend.clear()
        self.drift_calib_plot.setTitle('')
        # Get data:
        current_survey_name = self.comboBox_survey.currentText()
        survey = self.parent().campaign.surveys[current_survey_name]

        # Get list of selected stations:
        selected_station = self.comboBox_station.currentText()
        if selected_station == 'All':
            stations = [self.comboBox_station.itemText(i) for i in range(self.comboBox_station.count())]
            stations = stations[1:]  # Drop the first items ("All")
        else:
            stations = [self.comboBox_station.currentText()]

        if not stations:
            return

        # Calc. drift:
        method = self.comboBox_method.currentText()
        self.drift_poly_coeff = {}
        self.drift_start_time = None
        if method == 'numpy.polyfit':
            for station in stations:
                self.drift_poly_coeff[station] = survey.calc_drift_at_station_polyfit(station,
                                                                                  degree=self.spinBox_degree.value(),
                                                                                  obs_type='reduced',
                                                                                  active_only=self.checkBox_active_obs_only.isChecked(),
                                                                                  min_number_obs=self.spinBox_min_no_obs.value())
        else:
            raise RuntimeError(f'Unknown drift calculation method: {method}')

        # ### Plot data:
        # - Give stations different colours and show legend accordingly
        # - Mark/highlight non-active observations
        # self.station_color_dict = get_station_color_dict(stations, randomize=True)
        current_survey = self.parent().campaign.surveys[current_survey_name]
        obs_df = current_survey.obs_df.copy(deep=True)
        if obs_df is None:
            QMessageBox.warning(self, 'Warning!', f'No observation data available for survey {current_survey_name}')
            self.reset_gui()
            return

        # Timeline for drift plot
        # - Hours since survey start
        delta_t_min_h = 0.0 # = 0
        delta_t_max_h = (current_survey.end_time - current_survey.start_time) / np.timedelta64(1, 's') / 3600
        delta_t_h = np.linspace(delta_t_min_h, delta_t_max_h, settings.DRIFT_PLOT_NUM_ITEMS_IN_DRIFT_FUNCTION)
        # - Unix time
        t_ref_unix = dt.datetime.fromisoformat('1970-01-01T00:00:00')
        t_ref_unix = pytz.utc.localize(t_ref_unix)
        delta_t_unix_min = (current_survey.start_time - t_ref_unix) / np.timedelta64(1, 's')
        delta_t_unix_max = (current_survey.end_time - t_ref_unix) / np.timedelta64(1, 's')
        delta_t_unix = np.linspace(delta_t_unix_min, delta_t_unix_max,
                                     settings.DRIFT_PLOT_NUM_ITEMS_IN_DRIFT_FUNCTION)

        # Get gravity offset for plotting:
        if self.checkBox_remove_g_offset.isChecked():
            tmp_filter = obs_df['station_name'] == stations[0]
        else:
            tmp_filter = obs_df['station_name'].isin(stations)
        g_offset_mugal = round(obs_df.loc[tmp_filter, 'g_red_mugal'].mean() / 1000) * 1000  # integer mGal

        for station in stations:
            brush_color = self.station_color_dict[station]

            # Plot observations:
            tmp_filter = obs_df['station_name'] == station
            if self.checkBox_active_obs_only.isChecked():
                tmp_filter = tmp_filter & obs_df['keep_obs']
            obs_df_short = obs_df.loc[tmp_filter].copy(deep=True)
            # - Remove g offset w.r.t. the first station in the list:
            if self.checkBox_remove_g_offset.isChecked():
                if station != stations[0]:  # Apply offset w.r.t. the first station
                    g_mean_stat = obs_df_short['g_red_mugal'].mean()
                    g_offset_stat_mugal = g_mean_first_stat - g_mean_stat
                    # obs_df_short['g_red_mugal'] = obs_df_short['g_red_mugal'] + g_offset_stat
                elif station == stations[0]:
                    g_mean_first_stat = obs_df_short['g_red_mugal'].mean()
                    g_offset_stat_mugal = 0.0
            else:
                g_offset_stat_mugal = 0.0
            obs_df_short['t_ref_unix'] = (obs_df_short['obs_epoch'].values - np.datetime64('1970-01-01T00:00:00')) / np.timedelta64(1, 's')
            scatter = pg.ScatterPlotItem()
            spots = []
            # - prep. data for scatterplot:
            for index, row in obs_df_short.iterrows():
                if row['keep_obs']:
                    brush_color_scatter = brush_color
                    symbol_scatter = 'o'
                else:
                    brush_color_scatter = brush_color
                    symbol_scatter = 'x'
                spot_dic = {'pos': (row['t_ref_unix'], row['g_red_mugal'] - g_offset_mugal + g_offset_stat_mugal),
                    'size': 10,
                    'pen': {'color': 'k',
                            'width': 1},
                    'brush': brush_color_scatter,
                    'symbol': symbol_scatter}
                spots.append(spot_dic)
            scatter.addPoints(spots)
            self.drift_calib_plot.addItem(scatter)

            # Plot drift function:
            coeff_list = self.drift_poly_coeff[station]
            if coeff_list is None:  # Not enough ob. available to fit a polynomial
                legend_str = station
            else:
                yy_mugal = np.polyval(coeff_list, delta_t_h)
                pen = pg.mkPen(color='k', width=3)
                self.drift_calib_plot.plot(delta_t_unix, yy_mugal - g_offset_mugal + g_offset_stat_mugal,
                                           # name=f'{station}',
                                           pen=pen, symbol='d', symbolSize=4, symbolBrush=brush_color)
                if len(coeff_list) == 2:  # deg=1
                    legend_str = f'{station}: {coeff_list[0]/1000*24:0.4f} mGal/d'
                elif len(coeff_list) == 3:  # deg=2
                    legend_str = f'{station}: {coeff_list[0]/1000*(24**2):0.4f} mGal/d², {coeff_list[1]/1000*24:0.4f} mGal/d'
                elif len(coeff_list) == 4:  # deg=3
                    legend_str = f'{station}: {coeff_list[0]/1000*(24**3):0.4f} mGal/d³, {coeff_list[1]/1000*(24**2):0.4f} mGal/d², {coeff_list[2]/1000*24:0.4f} mGal/d'
                else:
                    legend_str = station

            # Add item to legend
            s_item_tmp = pg.ScatterPlotItem()
            s_item_tmp.setBrush(brush_color)
            s_item_tmp.setPen({'color': 'k', 'width': 1})
            s_item_tmp.setSize(10)
            self.drift_calib_plot.legend.addItem(s_item_tmp, legend_str)

        # Adjust plot window:
        self.drift_calib_plot.setLabel(axis='left', text=f'g [µGal] + {g_offset_mugal / 1000:.1f} mGal')
        # self.drift_calib_plot.setTitle(f'Drift function w.r.t. setup observations')
        self.drift_calib_plot.autoRange()

    def set_up_drift_plot_widget(self):
        """Set up `self.graphicsLayoutWidget_drift_plot` widget."""
        self.glw_drift_plot = self.graphicsLayoutWidget_drift_plot
        self.glw_drift_plot.setBackground('w')  # white background color
        self.drift_calib_plot = self.glw_drift_plot.addPlot(0, 0, name='ts_plot',
                                                axisItems={'bottom': TimeAxisItem(orientation='bottom')})
        # self.drift_calib_plot.setLabel(axis='bottom', text='Time')  # Takes too much space
        self.drift_calib_plot.addLegend()
        self.drift_calib_plot.showGrid(x=True, y=True)

    def save(self):
        """Save the drift results as log file and plot."""
        campaign_dir = self.parent().campaign.output_directory
        filename = self.comboBox_survey.currentText() + '_drift.txt'
        filename = os.path.join(self.parent().campaign.output_directory, filename)
        self.save_drift_file(filename)
        filename = self.comboBox_survey.currentText() + '_drift.png'
        filename = os.path.join(self.parent().campaign.output_directory, filename)
        self.save_drift_plot(filename)

    def save_drift_file(self, filename, verbose=True):
        """Save the drift calculation results to a file."""
        if verbose:
            print(f'Save drift results as {filename}')

        # Remove None entries from dict:
        poly_ceoffs = {}
        for key, value in self.drift_poly_coeff.items():
            if value is not None:
                poly_ceoffs[key] = value
        if len(poly_ceoffs) == 0:
            raise RuntimeError('No valid drift polynomials available for export to file.')

        poly_function_str = f'f(t) = a0(t0)'
        colnames = ['a0 [mGal]']
        for i in range(self.spinBox_degree.value()):
            poly_function_str = poly_function_str + f' + a{i + 1}*(t-t0)^{i + 1}'
            colnames.append(f'a{i + 1} [mGal/d^{i + 1}]')
        poly_df = pd.DataFrame.from_dict(poly_ceoffs, orient='index')
        poly_df = poly_df[poly_df.columns[::-1]]
        for colnum in range(poly_df.shape[1]):  # muGal/h => mGal/day
            poly_df.iloc[:, colnum] = poly_df.iloc[:, colnum] / 1000 * (24**colnum)
        poly_df.columns = colnames
        poly_df.loc[:, 'Station'] = poly_df.index
        colnames = ['Station'] + colnames
        poly_df = poly_df[colnames]
        poly_df_str = poly_df.to_string(index=False, float_format=lambda x: '{:.4f}'.format(x))
        start_time = self.parent().campaign.surveys[self.comboBox_survey.currentText()].start_time
        try:
            start_time_tz = start_time.tzname()
        except:
            start_time_tz = ''

        tmp_str = f'### Drift results for survey {self.comboBox_survey.currentText()} ###\n'
        tmp_str = tmp_str + f'Calculated by GravTools {__version__}\n'
        tmp_str = tmp_str + f'File created: {dt.datetime.now().strftime("%Y-%m-%d, %H:%M:%S")}\n'
        tmp_str = tmp_str + f'Method: {settings.DRIFT_CALC_METHODS[self.comboBox_method.currentText()]}\n'
        tmp_str = tmp_str + f'Polynomial degree: {self.spinBox_degree.value()}\n'
        tmp_str = tmp_str + f'Min. number of obs.: {self.spinBox_min_no_obs.value()}\n'
        tmp_str = tmp_str + f'Active obs. only: {self.checkBox_active_obs_only.isChecked()}\n'
        tmp_str = tmp_str + f'Multiple setups only: {self.checkBox_multiple_setups_only.isChecked()}\n'
        tmp_str = tmp_str + '\n'
        tmp_str = tmp_str + poly_function_str + '\n'
        tmp_str = tmp_str + f'... with [t]=days and [g]=mGal\n'
        if start_time_tz:
            tmp_str = tmp_str + f'... with t0 = {start_time.isoformat()} ({start_time_tz})\n'
        else:
            tmp_str = tmp_str + f'... with t0 = {start_time.isoformat()}\n'
        tmp_str = tmp_str + f'\n'
        tmp_str = tmp_str + poly_df_str + '\n'
        tmp_str = tmp_str + f'\n'

        if verbose:
            print(tmp_str)

        with open(filename, 'w') as f:
            f.write(tmp_str)

    def save_drift_plot(self, filename, verbose=True):
        """Save the current drift plot as png file."""
        # exporter = pg.exporters.ImageExporter(self.graphicsLayoutWidget_results_drift_plot.scene())
        exporter = pg.exporters.ImageExporter(self.graphicsLayoutWidget_drift_plot.scene())
        flag_export_successful = exporter.export(filename)
        if verbose and flag_export_successful:
            print(f'Save drift plot as {filename}')




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_setup_data.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog_setup_data(object):
    def setupUi(self, Dialog_setup_data):
        Dialog_setup_data.setObjectName("Dialog_setup_data")
        Dialog_setup_data.resize(436, 330)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(Dialog_setup_data)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.checkBox_drift_ref_epoch_active_obs_only = QtWidgets.QCheckBox(Dialog_setup_data)
        self.checkBox_drift_ref_epoch_active_obs_only.setObjectName("checkBox_drift_ref_epoch_active_obs_only")
        self.verticalLayout_3.addWidget(self.checkBox_drift_ref_epoch_active_obs_only)
        self.groupBox_obs_computation_method = QtWidgets.QGroupBox(Dialog_setup_data)
        self.groupBox_obs_computation_method.setObjectName("groupBox_obs_computation_method")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_obs_computation_method)
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioButton_variance_weighted_mean = QtWidgets.QRadioButton(self.groupBox_obs_computation_method)
        self.radioButton_variance_weighted_mean.setChecked(True)
        self.radioButton_variance_weighted_mean.setObjectName("radioButton_variance_weighted_mean")
        self.buttonGroup_computation_method = QtWidgets.QButtonGroup(Dialog_setup_data)
        self.buttonGroup_computation_method.setObjectName("buttonGroup_computation_method")
        self.buttonGroup_computation_method.addButton(self.radioButton_variance_weighted_mean)
        self.verticalLayout.addWidget(self.radioButton_variance_weighted_mean)
        self.radioButton_individual_obs = QtWidgets.QRadioButton(self.groupBox_obs_computation_method)
        self.radioButton_individual_obs.setObjectName("radioButton_individual_obs")
        self.buttonGroup_computation_method.addButton(self.radioButton_individual_obs)
        self.verticalLayout.addWidget(self.radioButton_individual_obs)
        self.verticalLayout_3.addWidget(self.groupBox_obs_computation_method)
        self.groupBox_sd_options = QtWidgets.QGroupBox(Dialog_setup_data)
        self.groupBox_sd_options.setToolTip("<html><head/><body><p>Options for the determination of standard deviations of setup observations.</p></body></html>")
        self.groupBox_sd_options.setObjectName("groupBox_sd_options")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_sd_options)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.radioButton_sd_from_obsfile = QtWidgets.QRadioButton(self.groupBox_sd_options)
        self.radioButton_sd_from_obsfile.setChecked(True)
        self.radioButton_sd_from_obsfile.setObjectName("radioButton_sd_from_obsfile")
        self.verticalLayout_2.addWidget(self.radioButton_sd_from_obsfile)
        self.radioButton_sd_default_per_obs = QtWidgets.QRadioButton(self.groupBox_sd_options)
        self.radioButton_sd_default_per_obs.setObjectName("radioButton_sd_default_per_obs")
        self.verticalLayout_2.addWidget(self.radioButton_sd_default_per_obs)
        self.radioButton_sd_defaul_per_setup = QtWidgets.QRadioButton(self.groupBox_sd_options)
        self.radioButton_sd_defaul_per_setup.setObjectName("radioButton_sd_defaul_per_setup")
        self.verticalLayout_2.addWidget(self.radioButton_sd_defaul_per_setup)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label_sd_default = QtWidgets.QLabel(self.groupBox_sd_options)
        self.label_sd_default.setEnabled(False)
        self.label_sd_default.setObjectName("label_sd_default")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_sd_default)
        self.spinBox_sd_default = QtWidgets.QSpinBox(self.groupBox_sd_options)
        self.spinBox_sd_default.setEnabled(False)
        self.spinBox_sd_default.setMinimum(1)
        self.spinBox_sd_default.setMaximum(999999)
        self.spinBox_sd_default.setProperty("value", 100)
        self.spinBox_sd_default.setObjectName("spinBox_sd_default")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.spinBox_sd_default)
        self.verticalLayout_2.addLayout(self.formLayout)
        self.verticalLayout_3.addWidget(self.groupBox_sd_options)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)

        self.retranslateUi(Dialog_setup_data)
        self.radioButton_sd_from_obsfile.toggled['bool'].connect(self.label_sd_default.setDisabled) # type: ignore
        self.radioButton_sd_from_obsfile.toggled['bool'].connect(self.spinBox_sd_default.setDisabled) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog_setup_data)

    def retranslateUi(self, Dialog_setup_data):
        _translate = QtCore.QCoreApplication.translate
        Dialog_setup_data.setWindowTitle(_translate("Dialog_setup_data", "Setup data options"))
        self.checkBox_drift_ref_epoch_active_obs_only.setToolTip(_translate("Dialog_setup_data", "Consider only active observations for the determination of the drift polynomial reference epoch(s)."))
        self.checkBox_drift_ref_epoch_active_obs_only.setText(_translate("Dialog_setup_data", "Consider only active observations for the reference epoch"))
        self.groupBox_obs_computation_method.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Select the method for the computation of setup observations.</p></body></html>"))
        self.groupBox_obs_computation_method.setTitle(_translate("Dialog_setup_data", "Setup observations computation method"))
        self.radioButton_variance_weighted_mean.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Computa setup observations as the variance weighted mean ob all consecutive observations within one setup. This is the default option for the adjustment of gravity surveys comprising multiple stations or the estimatuion of vertical gravity gradients.</p></body></html>"))
        self.radioButton_variance_weighted_mean.setText(_translate("Dialog_setup_data", "Variance weighted mean"))
        self.radioButton_individual_obs.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Trest each individual observation as setup observation, i.e. do aggregate individual observations at all. This option may be used for analysing  stationary long-term observations at a single station, e.g. for the analysis of tides of the instrument\'s drift behaviour.</p></body></html>"))
        self.radioButton_individual_obs.setText(_translate("Dialog_setup_data", "Individual observations"))
        self.groupBox_sd_options.setTitle(_translate("Dialog_setup_data", "Setup SD options"))
        self.radioButton_sd_from_obsfile.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Use the standard deviations (SD) provided in the observation files. The SD of setup observations are calculated according to the variance weighte mean gravity observations. </p></body></html>"))
        self.radioButton_sd_from_obsfile.setText(_translate("Dialog_setup_data", "Use SD from observation files"))
        self.radioButton_sd_default_per_obs.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Apply the default standard deviation (SD) to individual observations. The SD of setup observations are calculated according to the variance weighte mean gravity observations. Hence, setups with a higher number of individual observations get a lower SD according the <span style=\" font-style:italic;\">sqrt(n)</span> rule.</p></body></html>"))
        self.radioButton_sd_default_per_obs.setText(_translate("Dialog_setup_data", "Apply default SD to individual observations in setups"))
        self.radioButton_sd_defaul_per_setup.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Apply the default standard deviation (SD) to setup observations. Hence, the SD provided in the observation files are dismissed and all setup observations have equal weights, neglecting the number of repeated observations per setup.</p></body></html>"))
        self.radioButton_sd_defaul_per_setup.setText(_translate("Dialog_setup_data", "Apply default SD to setup observations"))
        self.label_sd_default.setToolTip(_translate("Dialog_setup_data", "Default standard deviation (SD) [µGal]. Min = 1 µGal, max. = 999999 µGal. "))
        self.label_sd_default.setText(_translate("Dialog_setup_data", "Default SD [µGal]"))
        self.spinBox_sd_default.setToolTip(_translate("Dialog_setup_data", "<html><head/><body><p>Default standard deviation (SD) [µGal]. Min = 1 µGal, max. = 999999 µGal.</p></body></html>"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_export_results.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog_export_results(object):
    def setupUi(self, Dialog_export_results):
        Dialog_export_results.setObjectName("Dialog_export_results")
        Dialog_export_results.resize(453, 745)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(Dialog_export_results)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.groupBox_general_settings = QtWidgets.QGroupBox(Dialog_export_results)
        self.groupBox_general_settings.setObjectName("groupBox_general_settings")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.groupBox_general_settings)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label_select_lsm_run = QtWidgets.QLabel(self.groupBox_general_settings)
        self.label_select_lsm_run.setObjectName("label_select_lsm_run")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_select_lsm_run)
        self.comboBox_select_lsm_run = QtWidgets.QComboBox(self.groupBox_general_settings)
        self.comboBox_select_lsm_run.setObjectName("comboBox_select_lsm_run")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.comboBox_select_lsm_run)
        self.label_export_path = QtWidgets.QLabel(self.groupBox_general_settings)
        self.label_export_path.setObjectName("label_export_path")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_export_path)
        self.label_export_comment = QtWidgets.QLabel(self.groupBox_general_settings)
        self.label_export_comment.setObjectName("label_export_comment")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_export_comment)
        self.label_export_path_show = QtWidgets.QLabel(self.groupBox_general_settings)
        self.label_export_path_show.setText("")
        self.label_export_path_show.setObjectName("label_export_path_show")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.label_export_path_show)
        self.label_export_comment_show = QtWidgets.QLabel(self.groupBox_general_settings)
        self.label_export_comment_show.setText("")
        self.label_export_comment_show.setObjectName("label_export_comment_show")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.label_export_comment_show)
        self.verticalLayout_5.addLayout(self.formLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.checkBox_add_lsm_comment_to_filename = QtWidgets.QCheckBox(self.groupBox_general_settings)
        self.checkBox_add_lsm_comment_to_filename.setChecked(True)
        self.checkBox_add_lsm_comment_to_filename.setObjectName("checkBox_add_lsm_comment_to_filename")
        self.horizontalLayout_2.addWidget(self.checkBox_add_lsm_comment_to_filename)
        self.verticalLayout_5.addLayout(self.horizontalLayout_2)
        self.verticalLayout_4.addWidget(self.groupBox_general_settings)
        self.groupBox_nsb_file = QtWidgets.QGroupBox(Dialog_export_results)
        self.groupBox_nsb_file.setObjectName("groupBox_nsb_file")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_nsb_file)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.checkBox_write_nsb_file = QtWidgets.QCheckBox(self.groupBox_nsb_file)
        self.checkBox_write_nsb_file.setChecked(True)
        self.checkBox_write_nsb_file.setObjectName("checkBox_write_nsb_file")
        self.verticalLayout_3.addWidget(self.checkBox_write_nsb_file)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.formLayout_3 = QtWidgets.QFormLayout()
        self.formLayout_3.setObjectName("formLayout_3")
        self.radioButton_first_dhb_dhf = QtWidgets.QRadioButton(self.groupBox_nsb_file)
        self.radioButton_first_dhb_dhf.setChecked(False)
        self.radioButton_first_dhb_dhf.setObjectName("radioButton_first_dhb_dhf")
        self.buttonGroup_dhb_dhf = QtWidgets.QButtonGroup(Dialog_export_results)
        self.buttonGroup_dhb_dhf.setObjectName("buttonGroup_dhb_dhf")
        self.buttonGroup_dhb_dhf.addButton(self.radioButton_first_dhb_dhf)
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.radioButton_first_dhb_dhf)
        self.radioButton_mean_dhb_dhf = QtWidgets.QRadioButton(self.groupBox_nsb_file)
        self.radioButton_mean_dhb_dhf.setChecked(True)
        self.radioButton_mean_dhb_dhf.setObjectName("radioButton_mean_dhb_dhf")
        self.buttonGroup_dhb_dhf.addButton(self.radioButton_mean_dhb_dhf)
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.radioButton_mean_dhb_dhf)
        self.checkBox_nsb_remove_datum_stations = QtWidgets.QCheckBox(self.groupBox_nsb_file)
        self.checkBox_nsb_remove_datum_stations.setChecked(True)
        self.checkBox_nsb_remove_datum_stations.setObjectName("checkBox_nsb_remove_datum_stations")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.checkBox_nsb_remove_datum_stations)
        self.radioButton_export_sd = QtWidgets.QRadioButton(self.groupBox_nsb_file)
        self.radioButton_export_sd.setObjectName("radioButton_export_sd")
        self.buttonGroup_sd_se = QtWidgets.QButtonGroup(Dialog_export_results)
        self.buttonGroup_sd_se.setObjectName("buttonGroup_sd_se")
        self.buttonGroup_sd_se.addButton(self.radioButton_export_sd)
        self.formLayout_3.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.radioButton_export_sd)
        self.radioButton_export_se = QtWidgets.QRadioButton(self.groupBox_nsb_file)
        self.radioButton_export_se.setChecked(True)
        self.radioButton_export_se.setObjectName("radioButton_export_se")
        self.buttonGroup_sd_se.addButton(self.radioButton_export_se)
        self.formLayout_3.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.radioButton_export_se)
        self.horizontalLayout.addLayout(self.formLayout_3)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.verticalLayout_4.addWidget(self.groupBox_nsb_file)
        self.groupBox_other_files = QtWidgets.QGroupBox(Dialog_export_results)
        self.groupBox_other_files.setObjectName("groupBox_other_files")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_other_files)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.checkBox_write_log_file = QtWidgets.QCheckBox(self.groupBox_other_files)
        self.checkBox_write_log_file.setToolTip("")
        self.checkBox_write_log_file.setChecked(True)
        self.checkBox_write_log_file.setObjectName("checkBox_write_log_file")
        self.verticalLayout_2.addWidget(self.checkBox_write_log_file)
        self.checkBox_save_drift_plot_png = QtWidgets.QCheckBox(self.groupBox_other_files)
        self.checkBox_save_drift_plot_png.setEnabled(True)
        self.checkBox_save_drift_plot_png.setChecked(True)
        self.checkBox_save_drift_plot_png.setObjectName("checkBox_save_drift_plot_png")
        self.verticalLayout_2.addWidget(self.checkBox_save_drift_plot_png)
        self.checkBox_save_vg_plot_png = QtWidgets.QCheckBox(self.groupBox_other_files)
        self.checkBox_save_vg_plot_png.setChecked(True)
        self.checkBox_save_vg_plot_png.setObjectName("checkBox_save_vg_plot_png")
        self.verticalLayout_2.addWidget(self.checkBox_save_vg_plot_png)
        self.verticalLayout_4.addWidget(self.groupBox_other_files)
        self.groupBox_observation_list = QtWidgets.QGroupBox(Dialog_export_results)
        self.groupBox_observation_list.setObjectName("groupBox_observation_list")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_observation_list)
        self.verticalLayout.setObjectName("verticalLayout")
        self.checkBox_write_observation_list = QtWidgets.QCheckBox(self.groupBox_observation_list)
        self.checkBox_write_observation_list.setChecked(True)
        self.checkBox_write_observation_list.setObjectName("checkBox_write_observation_list")
        self.verticalLayout.addWidget(self.checkBox_write_observation_list)
        self.formLayout_5 = QtWidgets.QFormLayout()
        self.formLayout_5.setObjectName("formLayout_5")
        spacerItem2 = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.formLayout_5.setItem(0, QtWidgets.QFormLayout.LabelRole, spacerItem2)
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.label_observation_list_export_options = QtWidgets.QLabel(self.groupBox_observation_list)
        self.label_observation_list_export_options.setObjectName("label_observation_list_export_options")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_observation_list_export_options)
        self.comboBox_observation_list_export_options = QtWidgets.QComboBox(self.groupBox_observation_list)
        self.comboBox_observation_list_export_options.setObjectName("comboBox_observation_list_export_options")
        self.comboBox_observation_list_export_options.addItem("")
        self.comboBox_observation_list_export_options.addItem("")
        self.comboBox_observation_list_export_options.addItem("")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.comboBox_observation_list_export_options)
        self.formLayout_5.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.formLayout_2)
        self.verticalLayout.addLayout(self.formLayout_5)
        self.verticalLayout_4.addWidget(self.groupBox_observation_list)
        self.groupBox_gis = QtWidgets.QGroupBox(Dialog_export_results)
        self.groupBox_gis.setObjectName("groupBox_gis")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.groupBox_gis)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.checkBox_gis_write_shapefile = QtWidgets.QCheckBox(self.groupBox_gis)
        self.checkBox_gis_write_shapefile.setObjectName("checkBox_gis_write_shapefile")
        self.verticalLayout_6.addWidget(self.checkBox_gis_write_shapefile)
        self.verticalLayout_4.addWidget(self.groupBox_gis)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem3)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog_export_results)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_4.addWidget(self.buttonBox)

        self.retranslateUi(Dialog_export_results)
        self.buttonBox.accepted.connect(Dialog_export_results.accept) # type: ignore
        self.buttonBox.rejected.connect(Dialog_export_results.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog_export_results)

    def retranslateUi(self, Dialog_export_results):
        _translate = QtCore.QCoreApplication.translate
        Dialog_export_results.setWindowTitle(_translate("Dialog_export_results", "Export data"))
        self.groupBox_general_settings.setToolTip(_translate("Dialog_export_results", "General export settings."))
        self.groupBox_general_settings.setTitle(_translate("Dialog_export_results", "General settings"))
        self.label_select_lsm_run.setText(_translate("Dialog_export_results", "Select LSM run"))
        self.comboBox_select_lsm_run.setToolTip(_translate("Dialog_export_results", "Select one LSM run for data export. If no LSM run is selected, just the observation list csv file can be exported, containing the current observation selection drom the observations tab."))
        self.label_export_path.setToolTip(_translate("Dialog_export_results", "Export path for all files. The path is defined when creating a new campaign and can be changed via the menu \"File/Change output directory\"."))
        self.label_export_path.setText(_translate("Dialog_export_results", "Export_path:"))
        self.label_export_comment.setToolTip(_translate("Dialog_export_results", "Comment of the selected LSM run."))
        self.label_export_comment.setText(_translate("Dialog_export_results", "Comment:"))
        self.checkBox_add_lsm_comment_to_filename.setToolTip(_translate("Dialog_export_results", "Append the LSM run comment to the name of the output file(s), e.g. <campaign name>_<lsm run comment>.nsb"))
        self.checkBox_add_lsm_comment_to_filename.setText(_translate("Dialog_export_results", "Append comment to filenames"))
        self.groupBox_nsb_file.setToolTip(_translate("Dialog_export_results", "Export options for nsb files that contain the estimation results (gravity at stations) of a network adjustment. This option is only available if station gravity was estimated (e.g. not for estimation of gravity gradients)"))
        self.groupBox_nsb_file.setTitle(_translate("Dialog_export_results", "nsb file"))
        self.checkBox_write_nsb_file.setToolTip(_translate("Dialog_export_results", "Write a nsb file."))
        self.checkBox_write_nsb_file.setText(_translate("Dialog_export_results", "Write nsb file"))
        self.radioButton_first_dhb_dhf.setToolTip(_translate("Dialog_export_results", "Take vertical offsets between instrument top and ground (dhb) and instrument top and reference point (dhf) at a station from the first setup at this station."))
        self.radioButton_first_dhb_dhf.setText(_translate("Dialog_export_results", "Take dhb and dhf from first setup at a station"))
        self.radioButton_mean_dhb_dhf.setToolTip(_translate("Dialog_export_results", "Calculate mean values of the vertical offsets between instrument top and ground (dhb) and instrument top and reference point (dhf) from all setups at a station. Write these mean values to the nsb file."))
        self.radioButton_mean_dhb_dhf.setText(_translate("Dialog_export_results", "Use mean dhb and dhf at stations"))
        self.checkBox_nsb_remove_datum_stations.setToolTip(_translate("Dialog_export_results", "Remove all datum stations rom the nsb file."))
        self.checkBox_nsb_remove_datum_stations.setText(_translate("Dialog_export_results", "Remove datum stations from nsb file"))
        self.radioButton_export_sd.setToolTip(_translate("Dialog_export_results", "<html><head/><body><p>Export the standard deviations (SD) of the gravity estimates.</p></body></html>"))
        self.radioButton_export_sd.setText(_translate("Dialog_export_results", "Export std. deviation (SD)"))
        self.radioButton_export_se.setToolTip(_translate("Dialog_export_results", "<html><head/><body><p>Export the standard errors (SE) of the gravity estimates.</p></body></html>"))
        self.radioButton_export_se.setText(_translate("Dialog_export_results", "Export std. error (SE)"))
        self.groupBox_other_files.setTitle(_translate("Dialog_export_results", "Other files"))
        self.checkBox_write_log_file.setText(_translate("Dialog_export_results", "Write log file (text file)"))
        self.checkBox_save_drift_plot_png.setToolTip(_translate("Dialog_export_results", "Save the current drift plot to a PNG file."))
        self.checkBox_save_drift_plot_png.setText(_translate("Dialog_export_results", "Save current drift plot view (PNG file)"))
        self.checkBox_save_vg_plot_png.setToolTip(_translate("Dialog_export_results", "Save the current VG plot to a PNG file (only available if the VG was estimated)."))
        self.checkBox_save_vg_plot_png.setText(_translate("Dialog_export_results", "Save the current VG plot view (PNG file)"))
        self.groupBox_observation_list.setToolTip(_translate("Dialog_export_results", "The observation list file contains either a list of all observations, or of all active or inactive observations. If a LSM run is selected, the observation list that was used to calculate the setup observations for this particular run will be exported. If no LSM run is selected, the current observation list that is visualized in the observations tab (obs. plot and obs. table) will be exported. Observation list files (csv) can be loaded by GravTools at any time to restore a particular selection of observations. "))
        self.groupBox_observation_list.setTitle(_translate("Dialog_export_results", "Observation list (CSV file)"))
        self.checkBox_write_observation_list.setToolTip(_translate("Dialog_export_results", "Write an observation list file."))
        self.checkBox_write_observation_list.setText(_translate("Dialog_export_results", "Write Observation list"))
        self.label_observation_list_export_options.setToolTip(_translate("Dialog_export_results", "Select which observations will be written to the file depending in the activity status."))
        self.label_observation_list_export_options.setText(_translate("Dialog_export_results", "Export "))
        self.comboBox_observation_list_export_options.setToolTip(_translate("Dialog_export_results", "Select which observations will be written to the file depending in the activity status."))
        self.comboBox_observation_list_export_options.setItemText(0, _translate("Dialog_export_results", "all observations"))
        self.comboBox_observation_list_export_options.setItemText(1, _translate("Dialog_export_results", "only active observations"))
        self.comboBox_observation_list_export_options.setItemText(2, _translate("Dialog_export_results", "only inactive observations"))
        self.groupBox_gis.setTitle(_translate("Dialog_export_results", "GIS data"))
        self.checkBox_gis_write_shapefile.setToolTip(_translate("Dialog_export_results", "<html><head/><body><p>Write the stations and observation results to an ESRI shapefile (content of the tabs observation and station results table). The coordinate reference system (crs) of the station coordinates has to be specified by an EPSG code!</p></body></html>"))
        self.checkBox_gis_write_shapefile.setText(_translate("Dialog_export_results", "Write station and observation results to a shapefile"))



"""Modelling gravity stations.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import pandas as pd
from gravtools.models.exceptions import FileTypeError
from gravtools.settings import STATION_DATA_SOURCE_TYPES


class Station:
    """Holds information on stations with known attributes.

    Attributes
    ----------
    station_files : dict
        The keys are the filenames (valid filename and path) and the values are the types of the station file
        (see: STATION_DATA_SOURCE_TYPES.keys()).
    stat_df : :py:obj:`pandas.core.frame.DataFrame`
        Pandas Dataframe that holds all station data. The columns are specified in :py:obj:`.Station._STAT_DF_COLUMNS`:

        - station_name : str
            Name of station.
        - long_deg : float
            Geographical longitude of the station [°] as optioned from the data source. Be aware that details on the
            used reference fame are not handled here!
        - lat_deg : float
            Geographical longitude of the station [°] as optioned from the data source. Be aware that details on the
            used reference fame are not handled here!
        - height_deg : float
            Height of the station [m] as optioned from the data source. Be aware that details on the
            used reference fame are not handled here!
        - g_mugal : float
            Gravity at the station [µGal] as optioned from the data source.
        - sd_g_mugal : float
            Standard deviation of the gravity at the station [µGal] as optioned from the data source.
        - vg_mugalm : float, optional (default=np.nan)
            Vertical gradient at the station. If `NaN`, ths value is not available (and will probably be replaced by
            the standard value defined in :py:obj:`gravtools.const.VG_DEFAULT`).
        - is_observed : bool
            Flag, that indicates whether the station was observed at least once in at least one survey in the current
            campaign. If `True`, the station was observed.
        - source_type : str
            The data source type indicates where the station data originates. It hast to be listed in
            :py:obj:`gravtools.settings.STATION_DATA_SOURCE_TYPES`.
        - is_datum : bool
            This flag indicates, whether this station is used as datum station in the analysis. Hence, this is an
            estimation setting. For stations from observation files the default is `False`, for ÖSGN stations the
            default is `True`.
        - in_survey : str
            String that lists the names of the surveys in which the station was observed in (separated by ';')
    """

    _STAT_DF_COLUMNS_DTYPES = {
        'station_name': 'str',  # Station name, str
        'long_deg': 'float',  # longitude [deg], float
        'lat_deg': 'float',  # latitude [deg], float
        'height_m': 'float',  # Height [m]
        'g_mugal': 'float',  # gravity [µGal]
        'sd_g_mugal': 'float',  # standard deviation of the gravity [µGal]
        'vg_mugalm': 'float',  # vertical gradient [µGal/m]
        'is_observed': 'bool',  # flag: True, if station was observed in at least one survey in the current campaign.
        'source_type': 'str',  # Source of the station data, str
        'source_name': 'str',  # Name of the data source
        'is_datum': 'bool',  # flag, that indicates whether this is a datum station, bool
        'in_survey': 'str',  # Names of surveys in which ths station was observed (separator = ;), str
    }
    _STAT_DF_COLUMNS = list(_STAT_DF_COLUMNS_DTYPES.keys())

    # _STAT_DF_COLUMNS = (
    #     'station_name',  # Station name, str
    #     'long_deg',  # longitude [deg], float
    #     'lat_deg',  # latitude [deg], float
    #     'height_m',  # Height [m]
    #     'g_mugal',  # gravity [µGal]
    #     'sd_g_mugal',  # standard deviation of the gravity [µGal]
    #     'vg_mugalm',  # vertical gradient [µGal/m]
    #     'is_observed',  # flag: True, if station was observed in at least one survey in the current campaign.
    #     'source_type',  # Source of the station data, str
    #     'source_name',  # Name of the data source
    #     'is_datum',  # flag, that indicates whether this is a datum station, bool
    #     'in_survey',  # Names of surveys in which ths station was observed (separator = ;), str
    # )

    def __init__(self, station_files=None):
        """
        Parameters
        ----------
        station_files : dict, optional
            The keys are the filenames (valid filename and path) and the values are the types of the station file
            (see: STATION_DATA_SOURCE_TYPES.keys()). Default=None which implies that no stations are added to this
            object.

        Raises
        ------
        FileTypeError
            Station file format not defined in :py:const:`gravtools.settings.STATION_DATA_SOURCE_TYPES`.
        TypeError
            Constructor arguments of wrong type.
        """
        if station_files is None:
            station_files = {}
        if not isinstance(station_files, dict):
            raise TypeError('"station_files" has to be a dict!')

        self.station_files = station_files

        # Init. station dataframe:
        self.stat_df = pd.DataFrame(columns=self._STAT_DF_COLUMNS)
        self.stat_df = self.stat_df.astype(self._STAT_DF_COLUMNS_DTYPES)

        # Loop over station_files and append the content to stat_df:
        for file_name, file_type in station_files.items():
            if file_type not in STATION_DATA_SOURCE_TYPES.keys():
                raise FileTypeError(f'File type of {file_name} not valid.',
                                    valid_file_types=[*STATION_DATA_SOURCE_TYPES])
            # Load file dependent on the file type:
            if file_type == 'oesgn_table':
                self.add_stations_from_oesgn_table(file_name)
            else:
                raise NotImplementedError(f'Loading station files of type "{file_type}" not implemented yet.')

    def _read_oesgn_table(self, filename, is_datum=False):
        """Reads an OESGN table file and returns a dataframe containing this data.

        Notes
        -----
        Currently it is only possible to load station data from the OESGN Table data file!

        Parameters
        ----------
        filename : str
            Name (and path) of the OESGN table file.
        is_datum : bool, optional (default = False)
            `True` indicates that all loaded OESGN stations are initially selected as datum stations (`is_datum`=True)

        Returns
        -------
        :py:obj:`pandas.core.frame.DataFrame`
            Pandas dataframe with all columns specified in :py:attr:`.Station._STAT_DF_COLUMNS` containing the data of
            the OESGN table given by the input argument `filename`.
        """
        widths = (
            10,  # Punktnummer im ÖSGN
            24,  # Anmerkung
            8,  # Geograph. Breite [deg] 34
            8,  # Geograph. Länge [deg] 42
            8,  # Höhe [m] 50
            7,  # g - 9.8e8 [µGal] 58
            3,  # mittlerer Fehler von g [µGal] 65
            4,  # Vertikalgradient der Schwere [µGal/m] 68
            6,  # Datum 73
            12,  # Punktidentität 79
        )
        column_names = (
            'station_name',
            'notes',
            'lat_deg',
            'long_deg',
            'height_m',
            'g_mugal',
            'sd_g_mugal',
            'vg_mugalm',
            'date',
            'identity'
        )

        columns_to_be_dropped = []
        for col_name in column_names:
            if col_name not in self._STAT_DF_COLUMNS:
                columns_to_be_dropped.append(col_name)

        try:
            stat_df_oesgn = pd.read_fwf(filename, widths=widths, header=None)
        except UnicodeDecodeError:
            # To catch encoding errors with files created/converted on Windows.
            # See: https://docs.python.org/3/library/codecs.html#standard-encodings
            # See: https://stackoverflow.com/questions/45529507/unicodedecodeerror-utf-8-codec-cant-decode-byte-0x96-in-position-35-invalid
            stat_df_oesgn = pd.read_fwf(filename, widths=widths, header=None, encoding='ISO-8859-1')

        stat_df_oesgn.columns = column_names
        stat_df_oesgn['is_observed'] = False  # Default = False
        if is_datum:
            stat_df_oesgn['is_datum'] = True
        else:
            stat_df_oesgn['is_datum'] = False
        stat_df_oesgn['source_type'] = 'oesgn_table'
        stat_df_oesgn['source_name'] = os.path.basename(filename)  # filename

        stat_df_oesgn.drop(columns=columns_to_be_dropped, inplace=True)

        stat_df_oesgn['height_m'] = stat_df_oesgn['height_m'] * 1e-3  # Conversion: [mm] => [m]

        return stat_df_oesgn

    def add_stations_from_oesgn_table(self, filename, is_datum=False, verbose=False):
        """Adds (appends) all stations of the input OESGN table to the station dataframe.

        Parameters
        ----------
        filename : str
            Name (and path) of the OESGN table file.
        is_datum : bool, optional (default = False)
            `True` indicates that all loaded OESGN stations are initially selected as datum stations (is_datum=`True`)
        verbose : bool, optional
            Print notifications, if `True` (default=`False`)
        """
        stat_df_new = self._read_oesgn_table(filename, is_datum=is_datum)
        stat_df_new = self._stat_df_add_columns(stat_df_new)
        stat_df_new = self._stat_df_reorder_columns(stat_df_new)
        self.add_stations(stat_df_new, data_source_type='oesgn_table', verbose=verbose)

    def delete_station(self, station_names: list, verbose=False):
        """
        Deletes all rows where the station name occurs in the `station_names` list.

        Parameters
        ----------
        station_names : list of str
            List of station names. The according station records will be deleted from `self.stat_df`.
        verbose : bool
            If `True`, print notification on deleted items.
        """
        idx_series = self.stat_df.station_name.isin(station_names)
        if verbose:
            deleted_df = self.stat_df[idx_series]
            print(f'Deleted {len(deleted_df)} rows in station dataframe:')
            for index, row in deleted_df.iterrows():
                print(f' - Station {row["station_name"]:10s} (row index {index:d})')
        self.stat_df = self.stat_df[~idx_series]

    @property
    def get_number_of_stations(self) -> int:
        """int : Returns the number of stations."""
        return len(self.stat_df)

    def __str__(self):
        return f"Station DataFrame containing {self.get_number_of_stations} stations."

    @property
    def get_all_stations(self):
        """:py:obj:`pandas.core.frame.DataFrame` : Returns all available stations."""
        return self.stat_df

    def set_observed_info_from_survey(self, survey):
        """Set the `is_observed` flags and the `in_survey` strings in the py:obj:`Station.stat_df` DataFrame.

        Parameters
        ----------
        survey : py:obj:`Survey`
            Survey object which is used as reference to set the `is_observed` flags in the stations Dataframe.
        """
        self.stat_df.loc[self.stat_df['station_name'].isin(survey.obs_df['station_name']), 'is_observed'] = True

        # Delete None entries:
        self.stat_df.loc[self.stat_df['station_name'].isin(survey.obs_df['station_name']) & self.stat_df[
            'in_survey'].isnull(), 'in_survey'] = ''
        # self.stat_df.loc[self.stat_df['station_name'].isin(survey.obs_df['station_name']), 'in_survey'] = ''

        # Add survey names:
        num_entries = len(
            self.stat_df.loc[self.stat_df['station_name'].isin(survey.obs_df['station_name']), 'in_survey'])
        self.stat_df.loc[self.stat_df['station_name'].isin(survey.obs_df['station_name']), 'in_survey'] = \
            self.stat_df.loc[self.stat_df['station_name'].isin(survey.obs_df['station_name']),
                             'in_survey'].str.cat([survey.name + '; '] * num_entries)

    @classmethod
    def _stat_df_add_columns(cls, stat_df):
        """Add and initialize (as None) all columns that are listed in cls._STAT_DF_COLUMNS and not present in input
        stat_df."""
        columns_to_be_initialized_as_none = list(set(cls._STAT_DF_COLUMNS) - set(stat_df.columns))
        stat_df[columns_to_be_initialized_as_none] = None
        return stat_df

    @classmethod
    def _stat_df_reorder_columns(cls, stat_df):
        """Change order of columns of stat_df to the order specified in cls._STAT_DF_COLUMNS.

        See: https://erikrood.com/Python_References/change_order_dataframe_columns_final.html
        """
        stat_df = stat_df[list(cls._STAT_DF_COLUMNS)]
        return stat_df

    @classmethod
    def _stat_df_check_columns(cls, stat_df, verbose=False) -> bool:
        """Check if all columns specified in cls._STAT_DF_COLUMNS are present."""
        is_valid = True

        invalid_cols = list(set(stat_df) - set(cls._STAT_DF_COLUMNS))
        if len(invalid_cols) > 0:
            is_valid = False
            if verbose:
                print(f'The following columns are not valid: {", ".join(invalid_cols)}')

        invalid_cols = list(set(cls._STAT_DF_COLUMNS) - set(stat_df))
        if len(invalid_cols) > 0:
            is_valid = False
            if verbose:
                print(f'The following columns are missing: {", ".join(invalid_cols)}')

        return is_valid

    def add_stations_from_survey(self, survey, verbose=False):
        """Add stations from a survey dataset, if they are not included in the station data yet.

        Notes
        -----
        The location information (longitude, latitude and height) are taken from the first appearance of a station in
        the survey dataframe, although this information may differ between setups if the location was determined (e.g.
        by GPS) for each individual measurement setup.

        Parameters
        ----------
        survey : py:obj:`Survey`
            Survey object which is search for stations that are not included in the station dataframe yet.
        verbose : bool, defualt=False
            If `True`, print notifications to command line.
        """
        # Get the essential information from the input survey:
        tmp_df = survey.obs_df[['station_name', 'lon_deg', 'lat_deg', 'alt_m', 'obs_epoch']].copy(deep=True)
        tmp_df = tmp_df.sort_values(by=['obs_epoch']).drop_duplicates(subset=['station_name']).drop(
            columns=['obs_epoch'])

        # Add missing columns (defined in self._STAT_DF_COLUMNS) to temp_df and initialize data:
        tmp_df.rename(columns={'lon_deg': 'long_deg', 'alt_m': 'height_m'}, inplace=True)
        tmp_df = self._stat_df_add_columns(tmp_df)
        tmp_df = self._stat_df_reorder_columns(tmp_df)
        tmp_df['is_datum'] = False
        tmp_df['is_observed'] = True
        tmp_df['source_type'] = 'obs_file'
        tmp_df['source_name'] = survey.name  # survey name
        tmp_df['in_survey'] = ''
        tmp_df.reset_index(drop=True, inplace=True)
        tmp_df = tmp_df.astype(self._STAT_DF_COLUMNS_DTYPES)

        self.add_stations(tmp_df, data_source_type='obs_file', verbose=verbose)

        # # Add missing stations to stat_df by matching the station names only (location parameters of observed stations
        # # may differ!):
        # self.stat_df = pd.concat([self.stat_df, tmp_df]).drop_duplicates(
        #     subset=['station_name'],
        #     keep='first'
        # ).reset_index(drop=True)

    def add_stations(self, stat_df_add, data_source_type, verbose=False):
        """Adds stations from various sources (e.g. station and observation files) to the station dataframe.

        Notes
        -----

        - Station names (column 'station_name') have to be unique in the resulting station dataframe
        - Keep the added station and drop existing entries, if the station names match.
        - If a station (with a unique station_name) is already available in the station dataframe and originates from a
            station file (e.g. ÖSGN file), the new station (from any source) is not added.
        - If a station originating from an observation file is already available in the station dataframe an the same
            station (same name) should be added from a station file (e.g. ÖSGN), the old version is overwritten be the
            entry from the station file.
        - If a station originating from an observation file is already available in the station dataframe and the same
            station (same name) should be added from an observation file (e.g. CG5 observation file), the old version
            is kept and the new version is discarded.

        Parameters
        ----------
        stat_df_add : py:oby:`.Station.stat_df`
            Station dataframe, that should be added so `self.stat_df`. The new station data has to come from one source
            (specified in `data_source_type`).
        data_source_type : str
            Specifies the type of the data source. Valid data source types are defined in
            :py:obj:`gravtools.settings.STATION_DATA_SOURCE_TYPES`.
        verbose : bool, default = `False`
            If `True`, the status messages are printed to the command line.

        """
        number_of_existing_stations = len(self.stat_df)

        # check, if the input station dataframe has valid columns:
        if not self._stat_df_check_columns(stat_df_add, verbose=verbose):
            raise AssertionError('The columns of the input dataframe are not valid!')

        # Check if data source type is valid:
        if data_source_type not in STATION_DATA_SOURCE_TYPES:
            raise AssertionError(f'The station data source type "{data_source_type}" is not valid!')

        # Check if the data in the input dataframe comes from the (valid) source given as input parameter:
        if not all(stat_df_add['source_type'] == data_source_type):
            raise AssertionError(f'All entries in the input station dataframe need to have the same data source '
                                 f'type ("{data_source_type}")!')

        # Add stations to a temp. copy of self.stat_df:
        stat_df_old = self.stat_df.copy(deep=True)

        # Add missing stations to stat_df by matching the station names only (location parameters of observed stations
        # may differ!):
        # - From station files => Drop existing entries with the same name in any case!
        if data_source_type == 'oesgn_table' or data_source_type == 'csv_stat_file':
            stat_df_new = pd.concat([stat_df_old, stat_df_add]).drop_duplicates(
                subset=['station_name'],
                keep='last'
            ).reset_index(drop=True)
        # From observation file  => Drop existing entries with the same name!
        elif data_source_type == 'obs_file':
            # => Replace entries with data_source = obs_file:
            stat_df_new = pd.concat([stat_df_old, stat_df_add]).drop_duplicates(
                subset=['station_name', 'source_type'],
                keep='last'
            ).reset_index(drop=True)
            # => Drop new entries from obs files, if entries from another data source type (station file) with the same
            # name already exist:
            stat_df_new = pd.concat([stat_df_new, stat_df_add]).drop_duplicates(
                subset=['station_name'],
                keep='first'
            ).reset_index(drop=True)

        self.stat_df = stat_df_new

        if verbose:
            number_of_new_stations = len(stat_df_add)
            number_of_stations = len(self.stat_df)
            stations_added = number_of_stations - number_of_existing_stations
            print(f"{number_of_new_stations} stations loaded. "
                  f"{stations_added} stations added ({number_of_new_stations - stations_added} already listed).")

    def add_stations_from_csv_file(self, filename, verbose=False):
        """Adds (appends) all stations of the input OESGN table to the station dataframe.

        Parameters
        ----------
        filename : str
            Name (and path) of the OESGN table file.
        verbose : bool, optional
            Print notifications, if `True` (default=`False`)
        """
        if verbose:
            print(f'Read station csv file: {filename}')
        stat_df_new = pd.read_csv(filename, comment='#')
        stat_df_new = self._stat_df_add_columns(stat_df_new)
        stat_df_new = self._stat_df_reorder_columns(stat_df_new)
        stat_df_new['is_observed'] = False  # Default = False
        stat_df_new['is_datum'] = False
        stat_df_new['source_type'] = 'csv_stat_file'
        stat_df_new['source_name'] = os.path.basename(filename)
        self.add_stations(stat_df_new, data_source_type='csv_stat_file', verbose=verbose)

    def set_datum_stations(self, station_names: list, is_datum: bool, verbose=False):
        """
        Dependent on the `is_datum` flag set the datum status of all stations in the list

        Parameters
        ----------
        station_names : list of str
            List of station names. The according station records will be deleted from `self.stat_df`.
        is_datum : bool
            `True` implies that all stations in the `station_names` list are set to datum stations. `False` implies
            that all stations in the list are no datum stations.
        verbose : bool
            If `True`, print notification to command line.
        """
        idx_series = self.stat_df.station_name.isin(station_names)
        if verbose:
            station_list = self.stat_df.loc[idx_series, 'station_name'].to_list()
            if is_datum:
                print(f'The following stations are defined as datum stations:')
            else:
                print(f'The following stations are NO datum stations:')
            for station_name in station_list:
                print(f' - {station_name}')
        self.stat_df.loc[idx_series, 'is_datum'] = is_datum



"""
Legacy processing pipeline for gravity observations at BEV, emulating the legacy fortran routines by Meurers & Ruess.
This module is part of the gravtools package.
Author: Andreas Hellerschmied
"""

import sys
import numpy as np

# Imports from other gravtools modules:
from bev_legacy import settings
from bev_legacy import init, drift_mlr, output, plots


def calc_abs_g(stat_df):
    """
    Berechnung absluter Schwerewerte bezogen auf das ÖSGN, ausgehend von Drift-korrigierten Lesungen eines
    Relativgravimeters. Für die Auswertung einzelner Tagesmessungen.
    Berechnung analog zu Fortran Code SCHWAUS2016.FOR von D. Ruess ()
    :param stat_df:
    :return: stat_df (mit Ergebnissen)
    """

    # ### Parameters: ###
    # Lesungsabhängiger Fehler (als konstant angenommen, unabhängig vom Betrag der Ablesung und vom Gravimeter-Typ):
    sig_reading = 2  # [µGal]

    # df, only containingÖSGN stations:
    df_stat_oesgn = stat_df.loc[stat_df['is_oesgn']].copy()  # Ensure to make a copy instead of a sliced view of stat_df

    # Differences between drift-corrected readings and OESGN g-values:
    df_stat_oesgn['g0_mugal'] = df_stat_oesgn['g_oesgn_mugal'] - df_stat_oesgn['g_est_mugal']
    df_stat_oesgn['sig_g0_mugal'] = np.sqrt(df_stat_oesgn['sig_g_est_mugal'] ** 2
                                            + df_stat_oesgn['g_sig_oesgn_mugal'] ** 2
                                            + sig_reading ** 2)

    # Calculation of the weighted mean of all go:
    df_stat_oesgn['go_p'] = 1 / (df_stat_oesgn['sig_g0_mugal'] ** 2)
    xm_mugal = sum(df_stat_oesgn['go_p'] * df_stat_oesgn['g0_mugal']) / sum(
        df_stat_oesgn['go_p'])  # Weighted mean of g0 values (weights: go_p)
    if df_stat_oesgn.shape[0] == 1:  # Only one OEGSN stations measured
        sig_xm_mugal = df_stat_oesgn['g_sig_oesgn_mugal']
    else:
        n = df_stat_oesgn.shape[0]  # Number of observed ÖSGN stations
        sig_xm_mugal = np.sqrt(((sum(df_stat_oesgn['g0_mugal'] ** 2) - ((sum(df_stat_oesgn['g0_mugal'])) ** 2 / n)) /
                                (n * (n - 1))) + (1 / (sum(df_stat_oesgn['go_p']))))

    # Calculation of the absolute g values at all observed stations:
    stat_df['g_abs_mugal'] = stat_df['g_est_mugal'] + xm_mugal
    stat_df['sig_g_abs_mugal'] = np.sqrt(sig_xm_mugal ** 2 + stat_df['sig_g_est_mugal'] ** 2)
    stat_df['verb_mugal'] = stat_df['g_abs_mugal'] - stat_df['g_oesgn_mugal']
    stat_df['g_abs_full_mugal'] = stat_df['g_abs_mugal'] + 9.8e8  # Get absolute gravity value at station [µGal]

    # Add g0 and sig_go to stat_df:
    stat_df.update(df_stat_oesgn[['g0_mugal', 'sig_g0_mugal']])

    return stat_df, xm_mugal, sig_xm_mugal


def drift_schwaus(obs_df, stat_df, polynomial_degree, instrument_id, name_obs_file, out_path='',
                  skalierung=None, date_str='', timezone_str=''):
    """
    Nachbau aller Brechnungen auf DRIFT2011.FOR und SCHWAUS2016.FOR.

    :param obs_dict:
    :param obs_df:
    :param stat_df:
    :return:
    """

    # ### Drift-Korrektur ###
    # Analog zu Fortran Code DRIFT2011.FOR
    # Berechnungen mittels multipler linearer Regression
    # - 1.) Polynom vom Grad n=1,2,3 an die Lesungen anpassen
    # - 2.) Drift-korrigierte Lesungen bestimmen

    # ### Polynom zur Korrektur der Instrumentendrift vom Grad n anpassen  ###
    if settings.VERBOSE:
        print('Schätzung der Drift mit Polynom vom Grad {}:'.format(polynomial_degree))
    results_dict = drift_mlr.calc_drift_corr_mlr(obs_df, stat_df, polynomial_degree)
    stat_df = results_dict['stat_df']
    if settings.VERBOSE:
        # Print results:
        print(' - Polynom Koeffizeinten (sig = {:5.2f} µGal):'.format(results_dict['pol_coef_sig_mugal']))
        for degree, value in results_dict['pol_coef'].items():
            print('    b{deg} = {value:9.5f} µGal/h^{deg} ({value2:9.5f} mGal/Tag^{deg})'.format(deg=degree, value=value, value2=value*24/1000))
        print(' - Korrigierte Gravimeter-Lesungen je Station:')
        for i, values in stat_df.iterrows():
            print('    {}: {:12.2f} µGal (sig = {:5.2f} µGal)'.format(values['punktnummer'], values['g_est_mugal'],
                                                                      values['sig_g_est_mugal']))

    # ### Schwere bezogen auf das ÖSGN brechnen ###
    # - Analog zu Fortan Code SCHWAUS2016.FOR
    # - Berechnete absolute Schwere an den beobachteten Stationen bezogen auf das Höheniveau des Festpunktes!
    if settings.VERBOSE:
        print('Berechnung absoluter Schwere-Werte:')
        print(' - Lagerung der Drift-korrigierten Lesungen auf {} ÖSGN Stationen'.format(stat_df[stat_df['is_oesgn']
                                                                                                 == True].shape[0]))
    stat_df, g0_mean_mugal, sig_g0_mean_mugal = calc_abs_g(stat_df)
    if settings.VERBOSE:
        # Print results:
        print(' - Ergebnisse:')
        for i, values in stat_df.iterrows():
            print('    {}: g = {:12.2f} µGal (sig = {:5.2f} µGal), verb. = {:7.2f} µGal'.format(values['punktnummer'],
                                                                                                values[
                                                                                                    'g_abs_full_mugal'],
                                                                                                values[
                                                                                                    'sig_g_abs_mugal'],
                                                                                                values['verb_mugal']))

    # ### Write file for NSDB input ###
    path_name_nsb_file = out_path + name_obs_file + '.nsb'
    if settings.VERBOSE:
        print('NSDB Input Datei schreiben ({})'.format(path_name_nsb_file))
    output.write_nsb_file(stat_df, instrument_id, path_name_nsb_file)

    # ### Create Plot ###
    path_name_drift_plot = out_path + name_obs_file
    if settings.VERBOSE:
        print('Drift-Plot erstellen')
        if settings.FLAG_SAVE_DRIFT_PLOT_PDF:
            print(' - Speichern unter: {}_drift.pdf'.format(path_name_drift_plot))

    plots.create_drift_plot(obs_df, stat_df, results_dict['pol_coef'],
                            save_pdf=settings.FLAG_SAVE_DRIFT_PLOT_PDF,
                            session_name=name_obs_file,
                            path_save_file=out_path)

    # ### Write protocol file ###
    if settings.FLAG_CREATE_SCHWAUS_PROTOCOL:
        if settings.VERBOSE:
            print('Berechnungsprotokoll erstellen ({})'.format(out_path + name_obs_file + '_prot.txt'))
        output.write_schwaus_protcol(obs_df=obs_df,
                                     stat_df=stat_df,
                                     pol_coef=results_dict['pol_coef'],
                                     pol_coef_sig_mugal=results_dict['pol_coef_sig_mugal'],
                                     session_name=name_obs_file,
                                     path_save_file=out_path,
                                     instrument_id=instrument_id,
                                     skalierung=skalierung,
                                     g0_mean_mugal=g0_mean_mugal,
                                     sig_g0_mean_mugal=sig_g0_mean_mugal)


def main(path_oesgn_table, name_oesgn_table, path_obs_file, name_obs_file, out_path):

    obs_df, stat_df, df_oesgn, obs_info_dict = init.read_and_prep_data(path_oesgn_table,
                                                                       name_oesgn_table,
                                                                       path_obs_file,
                                                                       name_obs_file)

    # ### DRIFT2011 und SCHWAUS2016 ###
    drift_schwaus(obs_df, stat_df,
                  polynomial_degree=obs_info_dict['polynomial_degree'],
                  instrument_id=obs_info_dict['instrument_id'],
                  name_obs_file=obs_info_dict['filename'],
                  out_path=out_path,
                  skalierung=obs_info_dict['skalierung'],
                  date_str=obs_info_dict['date_str'],
                  timezone_str=obs_info_dict['timezone_str'])

    print('...fertig!')



# Run as standalone program:
if __name__ == "__main__":

    if len(sys.argv) == 1:
        print('Eingangsparameter aus options.py bezogen (Default-Parameter).')
        path_oesgn_table = settings.PATH_OESGN_TABLE
        name_oesgn_table = settings.NAME_OESGN_TABLE
        path_obs_file = settings.PATH_OBS_FILE_BEV
        name_obs_file = settings.NAME_OBS_FILE_BEV
        out_path = settings.OUT_PATH

    elif len(sys.argv) == 2:  # Name of observation file as input argument
        name_obs_file = sys.argv[1]
        # Default-parameters from options-file:
        path_oesgn_table = settings.PATH_OESGN_TABLE
        name_oesgn_table = settings.NAME_OESGN_TABLE
        path_obs_file = settings.PATH_OBS_FILE_BEV
        out_path = settings.OUT_PATH

    else:
        print('Error: Invalid number of input arguments!')
        exit()

    # Run init module:
    main(path_oesgn_table,
         name_oesgn_table,
         path_obs_file,
         name_obs_file,
         out_path)

else:
    # not run as standalone program, but as module
    pass



"""Model classes for PyQt's model view architecture.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5 import QtGui
from PyQt5.QtWidgets import QMessageBox
import datetime as dt
import pandas as pd


NONE_REPRESENTATION_IN_TABLE_VIEW = ''  # Representation of None values in table views in the GUI


class ResultsObservationModel(QAbstractTableModel):
    """Model for displaying the observations-related results."""

    # Number of decimal pla
    _DECIMAL_PLACES_PER_FLOAT_COLUMN = {
        'g_diff_mugal': 1,
        'sd_g_diff_mugal': 1,
        'sd_g_diff_est_mugal': 1,
        'v_diff_mugal': 1,
        'sd_v_diff_mugal': 1,
        'sd_g_mugal': 1,
        'g_mugal': 1,
        'abw_mugal': 1,
        'delta_t_h': 3,
        'corr_drift_mugal': 1,
        'w_diff_mugal': 3,
        'r_diff_obs': 3,
        'ref_epoch_dt': 3,
        'w_obs_est_mugal': 3,  # LSM_non_diff
        'r_obs_est': 3,  # LSM_non_diff
        'v_obs_est_mugal': 1,  # LSM_non_diff
        'sd_v_obs_est_mugal': 1,  # LSM_non_diff
        'sd_g_obs_est_mugal': 1,  # LSM_non_diff
        'g_obs_mugal': 1,  # LSM_non_diff
        'sd_g_obs_mugal': 1,  # LSM_non_diff
    }

    # Columns that will be shown in the table view, if available in the data (Also defines the order of columns):
    # - This is also e pre-selection for data columns that can be plotted in the observation plots in the results tab
    # - keys: Actual names of the dataframe columns
    # - items: Header names for the Table View Widget
    _SHOW_COLUMNS_IN_TABLE_DICT = {
        'survey_name': 'Survey',  # LSM_non_diff, LSM_diff
        'station_name_from': 'Station from',  # LSM_diff
        'station_name_to': 'Station to',  # LSM_diff
        'station_name': 'Station',  # MLR, LSM_non_diff
        'ref_epoch_dt': 'Epoch UTC',  # LSM_non_diff, LSM_diff
        'delta_t_h': 'd_t [h]',  # LSM_non_diff, LSM_diff
        'g_mugal': 'g [µGal]',  # MLR
        'g_diff_mugal': 'g [µGal]',  # LSM_diff
        'g_obs_mugal': 'g [µGal]',  # LSM_non_diff
        'sd_g_mugal': 'SD [µGal]',  # MLR
        'sd_g_diff_mugal': 'SD [µGal]',  # LSM_diff
        'sd_g_obs_mugal': 'SD [µGal]',  # LSM_non_diff
        'sd_g_diff_est_mugal': 'SD_est [µGal]',  # LSM_diff
        'sd_g_obs_est_mugal': 'SD_est [µGal]',  # LSM_non_diff
        'v_diff_mugal': 'v [µGal]',  # LSM_diff
        'v_obs_est_mugal': 'v [µGal]',  # LSM_non_diff
        'sd_v_obs_est_mugal': 'SD_v [µGal]',  # LSM_non_diff
        'sd_v_diff_mugal': 'SD_v [µGal]',  # LSM_non_diff
        'w_diff_mugal': 'w',  # LSM_diff
        'w_obs_est_mugal': 'w',  # LSM_non_diff
        'r_diff_obs': 'r',  # LSM_diff
        'r_obs_est': 'r',  # LSM_non_diff
        'tau_test_result': 'Outlier test',  # LSM_non_diff, LSM_diff
        'corr_drift_mugal': 'Drift corr. [µGal]',  # MLR: Estimated drift correction
        'abw_mugal': 'abw [µGal]',  # MLR: Difference between drift-corrected instrument reading and the estimated station gravity
    }
    _SHOW_COLUMNS_IN_TABLE = list(_SHOW_COLUMNS_IN_TABLE_DICT.keys())  # Actual list of columns to be shown

    # List of columns that are shown in the simple mode of the GUI
    _SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI = [
        'survey_name',
        'station_name_from',
        'station_name_to',
        'station_name',
        'ref_epoch_dt',
        'delta_t_h',
        'g_mugal',
        'g_diff_mugal',
        'g_obs_mugal',
        'sd_g_mugal',
        'sd_g_diff_mugal',
        'sd_g_obs_mugal',
        'sd_g_diff_est_mugal',
        'sd_g_obs_est_mugal',
        'v_diff_mugal',
        'v_obs_est_mugal',
        'tau_test_result',
    ]

    # Columns that can be plotted. Column names (keys) and description (values):
    # - numerical data only!
    _PLOTABLE_DATA_COLUMNS = {
        # LSM_diff:
        'v_diff_mugal': 'Post-fit residuals [µGal]',  # LSM_diff
        'w_diff_mugal': 'w',  # LSM_diff
        'r_diff_obs': 'Redundancy components []',  # LSM_diff
        'sd_v_diff_mugal': 'SD of post-fit residuals [µGal]',  # LSM_diff
        'sd_g_diff_est_mugal': 'A posteriori SD of diff. obs. [µGal]',  # LSM_diff
        'g_diff_mugal': 'Differential observation [µGal]',  # LSM_diff
        'sd_g_diff_mugal': 'SD of differential observation [µGal]',  # LSM_diff
        # MLR:
        'g_mugal': 'Instrument reading [µGal]',  # MLR
        'sd_g_mugal': 'Standard deviation of the instrument reading [µGal]',  # MLR
        'corr_drift_mugal': 'Estimated drift correction [µGal]',  # MLR
        'abw_mugal': 'Drift-corrected reading minus estimated station gravity (not absolute) [µGal]',  # MLR
        # LSM_non_diff:
        'v_obs_est_mugal': 'Post-fit residuals [µGal]',  # LSM_non_diff
        'w_obs_est_mugal': 'Standardized post-fit residuals []',  # LSM_non_diff
        'r_obs_est': 'Redundancy components []',  # LSM_non_diff
        'sd_v_obs_est_mugal': 'SD of post-fit residuals [µGal]',  # LSM_non_diff
        'sd_g_obs_est_mugal': 'A posteriori SD of observation [µGal]',  # LSM_non_diff
        'g_obs_mugal': 'Non-differential observation [µGal]',  # LSM_non_diff
        'sd_g_obs_mugal': 'SD of non-differential observation [µGal]',  # LSM_non_diff
    }

    def __init__(self, lsm_runs):
        """Initialize the observation-results table view model.

        Parameters
        ----------
        lsm_runs : list of py.obj:`gravtools.lsm.LSM` objects
        """
        QAbstractTableModel.__init__(self)
        self._lsm_runs = []
        self._data = None  # Observations (or at subset of them) of the survey with the name `self._data_survey_name`
        self.load_lsm_runs(lsm_runs)
        self._lsm_run_index = None  # Name of the Survey that is currently represented by `self._data`
        self._data_column_names = None
        self.flag_gui_simple_mode = False

    def load_lsm_runs(self, lsm_runs: list):
        """Load adjustment results.

        Notes
        -----
        The data is assigned by reference, i.e. all changes in `_surveys` will propagate to the data origin.
        """
        self._lsm_runs = lsm_runs

    def update_view_model(self, lsm_run_index: int, station_name=None, survey_name=None, gui_simple_mode=False):
        """Update the `_data` DataFrame that hold the actual data that is displayed."""
        self.flag_gui_simple_mode = gui_simple_mode
        flag_error_init = False
        if lsm_run_index == -1:  # No data available => Invalid index => Reset model data
            flag_error_init = True
        else:
            try:
                results_obs_df = self._lsm_runs[lsm_run_index].get_results_obs_df
            except KeyError:
                QMessageBox.critical(self.parent(), 'Error!', f'LSM run with index "{lsm_run_index}" not found!')
            else:
                if results_obs_df is None:  # E.g. no results in campaign yet
                    flag_error_init = True
                else:
                    self._lsm_run_index = lsm_run_index

                    # Get list of columns to be depicted via the table model:
                    # - Keep order of itemms in `self._SHOW_COLUMNS_IN_TABLE`
                    results_obs_df_columns_set = frozenset(results_obs_df.columns)
                    table_model_columns = [x for x in self.get_table_columns if x in results_obs_df_columns_set]

                    # Apply filter:
                    if ((station_name is not None) or (survey_name is not None)) and (results_obs_df is not None):
                        tmp_filter = pd.Series([True] * len(results_obs_df))  # Init boolean filter series
                        column_names = results_obs_df.columns

                        if station_name is not None:  # Filer data for station names
                            flag_is_diff_obs = ('station_name_from' in column_names) and ('station_name_to' in column_names)

                            # Differential observations => Check columns 'station_name_from' and 'station_name_to'
                            if flag_is_diff_obs:
                                tmp_filter = tmp_filter & ((results_obs_df['station_name_from'] == station_name) |
                                                           (results_obs_df['station_name_to'] == station_name))
                            else:  # No differential observations => Check column 'station_name'
                                tmp_filter = tmp_filter & (results_obs_df['station_name'] == station_name)

                        if survey_name is not None:  # Filer data for survey name
                            tmp_filter = tmp_filter & (results_obs_df['survey_name'] == survey_name)
                        try:
                            self._data = results_obs_df.loc[tmp_filter, table_model_columns].copy(deep=True)
                        except:  # Just in case any problem occurs...
                            self._data = results_obs_df.loc[tmp_filter, :].copy(deep=True)  # Show all columns
                    else:  # No filter
                        try:
                            self._data = results_obs_df.loc[:, table_model_columns].copy(deep=True)
                        except:  # Just in case any problem occurs...
                            self._data = results_obs_df.copy(deep=True)
                    self._data_column_names = self._data.columns.to_list()  # Show all columns
        if flag_error_init:
            self._data = None
            self._lsm_run_index = None
            self._data_column_names = None

    def rowCount(self, parent=None):
        if self._data is not None:
            return self._data.shape[0]
        else:
            return 0

    def columnCount(self, parent=None):
        if self._data is not None:
            return self._data.shape[1]
        else:
            return 0

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole or role == Qt.UserRole:
                value = self._data.iloc[index.row(), index.column()]
                column_name = self._data_column_names[index.column()]
                # Custom formatter (string is expected as return type):
                if value is None:  #
                    return NONE_REPRESENTATION_IN_TABLE_VIEW
                elif isinstance(value, float):
                    if value != value:  # True, if value is "NaN"
                        return NONE_REPRESENTATION_IN_TABLE_VIEW
                    else:
                        if column_name in self._DECIMAL_PLACES_PER_FLOAT_COLUMN.keys():
                            num_dec_places = self._DECIMAL_PLACES_PER_FLOAT_COLUMN[column_name]
                            return '{1:.{0}f}'.format(num_dec_places, value)
                        else:
                            return str(value)
                elif isinstance(value, dt.datetime):
                    return value.strftime("%Y-%m-%d, %H:%M:%S")
                else:  # all other
                    return str(value)

            if role == Qt.TextAlignmentRole:
                value = self._data.iloc[index.row(), index.column()]
                if isinstance(value, int) or isinstance(value, float):
                    # Align right, vertical middle.
                    return Qt.AlignVCenter + Qt.AlignRight

            if role == Qt.BackgroundRole:
                if 'tau_test_result' in self._data_column_names:
                    if index.column() == self._data_column_names.index('tau_test_result'):
                        value = self._data.iloc[index.row(), index.column()]
                        if value == 'passed':
                            return QtGui.QColor('green')
                        elif value == 'failed':
                            return QtGui.QColor('red')
        return None

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self._SHOW_COLUMNS_IN_TABLE_DICT[str(self._data.columns[section])]
            if orientation == Qt.Vertical:
                return str(self._data.index[section])

    def get_plotable_columns(self) -> dict:
        """Returns a dict with names and description of the plotable numerical columns.

        Notes
        -----
        If the model data is empty return an empty dict.
        """
        plotable_columns_dict = {}
        if self._data_column_names is not None:  # Data model is not empty
            for key, item in self._PLOTABLE_DATA_COLUMNS.items():
                # Check if key is a column-name in the current data
                if key in self._data_column_names:
                    plotable_columns_dict[key] = item
        return plotable_columns_dict

    @property
    def get_model_data_df(self):
        """Returns the model data dataframe."""
        return self._data

    @property
    def get_model_data_df_for_plotting(self):
        """Return all plotable rows on the model data dataframe.

        Notes
        -----
        Data is plotable, if the data can be referred station(s) and to a specific reference epoch! E.g. pseudo
        observations introduced for datum constraints are not plotable!
        """
        filter_tmp = self._data['is_constraint'] == False
        # Add further options to thze filter series, is needed!
        return self._data.loc[filter_tmp, :]

    @property
    def get_table_columns(self):
        """Returns a list with all columns names of the dataframe that should be copied to the view model."""
        if self.flag_gui_simple_mode:
            return [value for value in self._SHOW_COLUMNS_IN_TABLE if value in self._SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI]
        else:
            return self._SHOW_COLUMNS_IN_TABLE

    def get_short_column_description(self, column_name: str) -> str:
        """Returns the short description of the model column."""
        try:
            return self._SHOW_COLUMNS_IN_TABLE_DICT[column_name]
        except AttributeError:
            return ''



"""Model classes for PyQt's model view architecture.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5 import QtGui
from PyQt5.QtWidgets import QMessageBox
import datetime as dt

NONE_REPRESENTATION_IN_TABLE_VIEW = ''  # Representation of None values in table views in the GUI


class SurveyTableModel(QAbstractTableModel):
    """Model for displaying the surveys in a table view (QTableView)."""

    # Columns that will be shown in the table view:
    # - keys: Actual attribute names of the Survey object
    # - items: Header names for the Table View Widget
    _SHOW_ATTRIBUTES_IN_TABLE_DICT = {
        'name': 'Name',
        'keep_survey': 'Use data',
        'date': 'Date',
        'start_time_hhmmss_str': 'Start time',
        'end_time_hhmmss_str': 'End time',
        'duration_hhmmss_str': 'Duration',
        'operator': 'Operator',
        'institution': 'Institution',
        'gravimeter_type': 'Gravimeter',
        'gravimeter_serial_number': 'S/N',
        'number_of_setups': 'Setups',
        'number_of_observations': 'Obs.',
        'number_of_active_observations': 'Active obs.',
        'observed_stations_str': 'Stations',
        'data_file_name': 'Filename',
        'data_file_type': 'Filetype',
    }
    # _ATTRIBUTE_NAMES = list(_SHOW_ATTRIBUTES_IN_TABLE_DICT.keys())  # Actual list of columns to be shown

    # List of columns that are shown in the simple mode of the GUI
    _SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI = [
        # 'name',
        'keep_survey',
        'date',
        'operator',
        'institution',
        'gravimeter_type',
        'gravimeter_serial_number',
        'number_of_setups',
        'number_of_observations',
        'number_of_active_observations',
        'start_time_hhmmss_str',
        'end_time_hhmmss_str',
        'duration_hhmmss_str',
        'observed_stations_str',
    ]

    _ATTRIBUTES_TOOLTIPS = {
        # 'name': 'Survey name',
        'keep_survey': '`True` implies that the observations of this survey are used to calculate setup observations.',
        'date': 'Date',
        'operator': 'Operator',
        'institution': 'Institution',
        'gravimeter_type': 'Gravimeter type',
        'gravimeter_serial_number': 'Serial number of the gravimeter',
        'number_of_setups': 'Number of setups in this survey',
        'number_of_observations': 'Total number of observations',
        'number_of_active_observations': 'Number of active observations',
        'start_time_hhmmss_str': 'Start time',
        'end_time_hhmmss_str': 'End time',
        'duration_hhmmss_str': 'Duration',
        'observed_stations_str': 'Stations observed in the survey',
        'data_file_name': 'Name of the survey data file (data source)',
        'data_file_type': 'Type of the survey data file',
    }

    # Number of decimal pla
    _DECIMAL_PLACES_PER_FLOAT_COLUMN = {
        'number_of_setups': 0,
        'number_of_observations': 0,
    }

    def __init__(self, surveys: dict, gui_simple_mode=False):
        QAbstractTableModel.__init__(self)
        self._show_attributes_in_table_dict = {}
        self._survey_name_list = []
        self.flag_gui_simple_mode = gui_simple_mode
        self._data = surveys
        self._survey_name_list = list(surveys.keys())
        if not gui_simple_mode:
            self._show_attributes_in_table_dict = self._SHOW_ATTRIBUTES_IN_TABLE_DICT
        else:
            for key, value in self._SHOW_ATTRIBUTES_IN_TABLE_DICT.items():
                if key in self._SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI:
                    self._show_attributes_in_table_dict[key] = value
        self._attribute_names = list(self._show_attributes_in_table_dict.keys())  # Actual list of columns to be shown

    def rowCount(self, parent=None):
        return len(self._data)  # Number of surveys

    def columnCount(self, parent=None):
        return len(self._attribute_names)

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            survey_name = self._survey_name_list[index.row()]
            attribute_name = self._attribute_names[index.column()]
            try:
                value = getattr(self._data[survey_name], attribute_name)
            except AttributeError:
                value = 'n.a.'  # Not available

            # if role == Qt.DisplayRole:
            if role == Qt.DisplayRole or role == Qt.UserRole:

                # Draw checkboxes only in the "is_datum" column:
                if role == Qt.DisplayRole:
                    if attribute_name == 'keep_survey':
                        return ''

                # Custom formatter (string is expected as return type):
                if value is None:  #
                    return NONE_REPRESENTATION_IN_TABLE_VIEW
                elif isinstance(value, float):
                    if value != value:  # True, if value is "NaN"
                        return NONE_REPRESENTATION_IN_TABLE_VIEW
                    else:
                        if attribute_name in self._DECIMAL_PLACES_PER_FLOAT_COLUMN.keys():
                            num_dec_places = self._DECIMAL_PLACES_PER_FLOAT_COLUMN[attribute_name]
                            return '{1:.{0}f}'.format(num_dec_places, value)
                        else:
                            return str(value)
                elif isinstance(value, dt.date):
                    return value.strftime(value.strftime('%Y-%m-%d'))

                else:  # all other
                    return str(value)

            if role == Qt.TextAlignmentRole:
                # value = self._data.iloc[index.row(), index.column()]
                if isinstance(value, int) or isinstance(value, float):
                    # Align right, vertical middle.
                    return Qt.AlignVCenter + Qt.AlignRight

            if role == Qt.BackgroundRole:
                if not self._data[survey_name].keep_survey:
                    return QtGui.QColor('red')

                if attribute_name == 'number_of_active_observations':
                    if self._data[survey_name].number_of_active_observations == 0:
                        return QtGui.QColor('red')

            if role == Qt.CheckStateRole:
                try:
                    if attribute_name == 'keep_survey':
                        keep_obs_flag = self._data[survey_name].keep_survey
                        if keep_obs_flag:
                            return Qt.Checked
                        else:
                            return Qt.Unchecked
                except Exception:
                    return None
        return None

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if self._data is not None:
                if orientation == Qt.Horizontal:
                    attribute_name = self._attribute_names[section]
                    return self._show_attributes_in_table_dict[attribute_name]
                if orientation == Qt.Vertical:
                    return self._survey_name_list[section]

        if role == Qt.ToolTipRole:
            if orientation == Qt.Horizontal:
                attribute_name = self._attribute_names[section]
                try:
                    return self._ATTRIBUTES_TOOLTIPS[attribute_name]
                except KeyError:
                    return ''

    def setData(self, index, value, role):
        """Example: https://www.semicolonworld.com/question/58510/how-to-display-a-pandas-data-frame-with-pyqt5"""
        if not index.isValid():
            return False

        if role == Qt.CheckStateRole:
            survey_name = self._survey_name_list[index.row()]
            attribute_name = self._attribute_names[index.column()]
            if attribute_name == 'keep_survey':
                if value == Qt.Unchecked:
                    self._data[survey_name].keep_survey = False
                elif value == Qt.Checked:
                    self._data[survey_name].keep_survey = True
                else:
                    QMessageBox.warning(self.parent(), 'Warning!',
                                        f'Invalid value fpr keep observation flag: "{value}"')
                    return False
                index_left = index.siblingAtColumn(0)
                index_right = index.siblingAtColumn(self.columnCount() - 1)
                self.dataChanged.emit(index_left, index_right)  # Update the row

            return True  # Data successfully set

        return False

    def flags(self, index):
        """Enable editing of table items."""
        flags = super(self.__class__, self).flags(index)
        flags |= Qt.ItemIsSelectable
        flags |= Qt.ItemIsEnabled
        flags |= Qt.ItemIsDragEnabled
        flags |= Qt.ItemIsDropEnabled
        if index.column() == self._attribute_names.index('keep_survey'):
            # flags |= Qt.ItemIsEditable  # Use checkbox only!
            flags |= Qt.ItemIsUserCheckable
        return flags

    @property
    def get_data(self):
        return self._data

    def row_index_of_survey(self, survey_name: str):
        """Returns the row index of a survey in the current model"""
        if self.no_data:
            return None
        else:
            return self._survey_name_list.index(survey_name)

    def emit_data_changed_survey(self, survey_name: str):
        """Emits the data changed signal for the row of specifies survey."""
        if not self.no_data:
            index_left = self.index(self.row_index_of_survey(survey_name), 0)
            index_right = self.index(self.row_index_of_survey(survey_name), self.columnCount() - 1)
            self.dataChanged.emit(index_left, index_right)

    @property
    def no_data(self):
        return not self._data

    def get_survey_name_by_row_index(self, row_idx: int) -> str:
        """Returns the name of the survey displayed in row with the index `idx`."""
        return self._survey_name_list[row_idx]




"""
GravTools - A software toolbox for processing relative gravity surveys
======================================================================

Script files for GravTools.

License, contact and authorship
-------------------------------

Copyright (C) 2021  Andreas Hellerschmied

GravTools was developed at Austria's the *Federal Office of Metrology and Surveying (BEV)*.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

:Authors:
    Andreas Hellerschmied (andreas.hellerschmied@bev.gv.at)
"""


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_load_tsf_file.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DialogLoadTsfFile(object):
    def setupUi(self, DialogLoadTsfFile):
        DialogLoadTsfFile.setObjectName("DialogLoadTsfFile")
        DialogLoadTsfFile.resize(623, 385)
        DialogLoadTsfFile.setToolTip("")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(DialogLoadTsfFile)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.pushButton_select_file = QtWidgets.QPushButton(DialogLoadTsfFile)
        self.pushButton_select_file.setMaximumSize(QtCore.QSize(100, 16777215))
        self.pushButton_select_file.setObjectName("pushButton_select_file")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.pushButton_select_file)
        self.lineEdit_filename = QtWidgets.QLineEdit(DialogLoadTsfFile)
        self.lineEdit_filename.setObjectName("lineEdit_filename")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_filename)
        self.label_survey_name = QtWidgets.QLabel(DialogLoadTsfFile)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_survey_name.sizePolicy().hasHeightForWidth())
        self.label_survey_name.setSizePolicy(sizePolicy)
        self.label_survey_name.setMaximumSize(QtCore.QSize(100, 16777215))
        self.label_survey_name.setObjectName("label_survey_name")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_survey_name)
        self.lineEdit_survey_name = QtWidgets.QLineEdit(DialogLoadTsfFile)
        self.lineEdit_survey_name.setObjectName("lineEdit_survey_name")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lineEdit_survey_name)
        self.verticalLayout_3.addLayout(self.formLayout_2)
        self.groupBox_filter = QtWidgets.QGroupBox(DialogLoadTsfFile)
        self.groupBox_filter.setObjectName("groupBox_filter")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_filter)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(self.groupBox_filter)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.lineEdit_filter_location = QtWidgets.QLineEdit(self.groupBox_filter)
        self.lineEdit_filter_location.setObjectName("lineEdit_filter_location")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_filter_location)
        self.label_3 = QtWidgets.QLabel(self.groupBox_filter)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.lineEdit_filter_instrument = QtWidgets.QLineEdit(self.groupBox_filter)
        self.lineEdit_filter_instrument.setObjectName("lineEdit_filter_instrument")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lineEdit_filter_instrument)
        self.label_4 = QtWidgets.QLabel(self.groupBox_filter)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.lineEdit_filter_data_type = QtWidgets.QLineEdit(self.groupBox_filter)
        self.lineEdit_filter_data_type.setObjectName("lineEdit_filter_data_type")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.lineEdit_filter_data_type)
        self.verticalLayout_2.addLayout(self.formLayout)
        self.verticalLayout_3.addWidget(self.groupBox_filter)
        self.groupBox_options = QtWidgets.QGroupBox(DialogLoadTsfFile)
        self.groupBox_options.setObjectName("groupBox_options")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_options)
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioButton_effect = QtWidgets.QRadioButton(self.groupBox_options)
        self.radioButton_effect.setChecked(True)
        self.radioButton_effect.setObjectName("radioButton_effect")
        self.buttonGroup_effect_or_correction = QtWidgets.QButtonGroup(DialogLoadTsfFile)
        self.buttonGroup_effect_or_correction.setObjectName("buttonGroup_effect_or_correction")
        self.buttonGroup_effect_or_correction.addButton(self.radioButton_effect)
        self.verticalLayout.addWidget(self.radioButton_effect)
        self.radioButton_correction = QtWidgets.QRadioButton(self.groupBox_options)
        self.radioButton_correction.setObjectName("radioButton_correction")
        self.buttonGroup_effect_or_correction.addButton(self.radioButton_correction)
        self.verticalLayout.addWidget(self.radioButton_correction)
        self.checkBox_overwrite_channel = QtWidgets.QCheckBox(self.groupBox_options)
        self.checkBox_overwrite_channel.setChecked(True)
        self.checkBox_overwrite_channel.setObjectName("checkBox_overwrite_channel")
        self.verticalLayout.addWidget(self.checkBox_overwrite_channel)
        self.verticalLayout_3.addWidget(self.groupBox_options)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(DialogLoadTsfFile)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_3.addWidget(self.buttonBox)

        self.retranslateUi(DialogLoadTsfFile)
        self.buttonBox.accepted.connect(DialogLoadTsfFile.accept) # type: ignore
        self.buttonBox.rejected.connect(DialogLoadTsfFile.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(DialogLoadTsfFile)

    def retranslateUi(self, DialogLoadTsfFile):
        _translate = QtCore.QCoreApplication.translate
        DialogLoadTsfFile.setWindowTitle(_translate("DialogLoadTsfFile", "Load TSF file"))
        self.pushButton_select_file.setToolTip(_translate("DialogLoadTsfFile", "Select TSF file"))
        self.pushButton_select_file.setText(_translate("DialogLoadTsfFile", "Select TSF file"))
        self.lineEdit_filename.setToolTip(_translate("DialogLoadTsfFile", "Filename and path of the TSF file"))
        self.label_survey_name.setToolTip(_translate("DialogLoadTsfFile", "Name of the survey for which the time series data is loaded from the TSF file"))
        self.label_survey_name.setText(_translate("DialogLoadTsfFile", "Survey name"))
        self.lineEdit_survey_name.setToolTip(_translate("DialogLoadTsfFile", "Name of the survey for which the time series data is loaded from the TSF file"))
        self.groupBox_filter.setToolTip(_translate("DialogLoadTsfFile", "<html><head/><body><p>Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <span style=\" font-style:italic;\">&lt;location&gt;:&lt;instrument&gt;:&lt;data type&gt;</span>.</p></body></html>"))
        self.groupBox_filter.setTitle(_translate("DialogLoadTsfFile", "Filter"))
        self.label_2.setToolTip(_translate("DialogLoadTsfFile", "Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <location>:<instrument>:<data type>."))
        self.label_2.setText(_translate("DialogLoadTsfFile", "Location:"))
        self.lineEdit_filter_location.setToolTip(_translate("DialogLoadTsfFile", "Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <location>:<instrument>:<data type>."))
        self.label_3.setToolTip(_translate("DialogLoadTsfFile", "Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <location>:<instrument>:<data type>."))
        self.label_3.setText(_translate("DialogLoadTsfFile", "Instrument:"))
        self.lineEdit_filter_instrument.setToolTip(_translate("DialogLoadTsfFile", "Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <location>:<instrument>:<data type>."))
        self.label_4.setToolTip(_translate("DialogLoadTsfFile", "Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <location>:<instrument>:<data type>."))
        self.label_4.setText(_translate("DialogLoadTsfFile", "Data type:"))
        self.lineEdit_filter_data_type.setToolTip(_translate("DialogLoadTsfFile", "Filter the channels that are loaded from the TSF file by defining restrictions for the location, the instrumt and/or the data type. The filters are applied on the channel names in the [CHANNELS] block that have to follow the convention <location>:<instrument>:<data type>."))
        self.groupBox_options.setToolTip(_translate("DialogLoadTsfFile", "General options"))
        self.groupBox_options.setTitle(_translate("DialogLoadTsfFile", "Options"))
        self.radioButton_effect.setToolTip(_translate("DialogLoadTsfFile", "<html><head/><body><p>The channel data represents the gravity effect of phenomena, e.g. the gravitational attraction of sun and moon. Effects are <span style=\" font-weight:600;\">subtracted</span> from gravity observations in order to correct them (default for synthetic gravity calculated in TSoft).</p><p>See <span style=\" font-style:italic;\">TSoft manual (version 2.2.4 Release date 2015-09-09), p. 15</span>: &quot;<span style=\" font-style:italic;\">The loading calculation by Tsoft will result in the correction, not the effect. On the other hand the prediction of the solid Earth tides using the WDD parameter set provided by Tsoft directly will result in the effect. Therefore both must be treated with different sign in order to reduce a gravity time series correctly.</span>&quot;</p></body></html>"))
        self.radioButton_effect.setText(_translate("DialogLoadTsfFile", "Time series describe the gravity effect of phenomena"))
        self.radioButton_correction.setToolTip(_translate("DialogLoadTsfFile", "<html><head/><body><p>The channel data represents corrections for gravity observations, i.e. these values are <span style=\" font-weight:600;\">added</span> to gravity observations in order to correct them.</p><p>See <span style=\" font-style:italic;\">TSoft manual (version 2.2.4 Release date 2015-09-09), p. 15</span>: &quot;<span style=\" font-style:italic;\">The loading calculation by Tsoft will result in the correction, not the effect. On the other hand the prediction of the solid Earth tides using the WDD parameter set provided by Tsoft directly will result in the effect. Therefore both must be treated with different sign in order to reduce a gravity time series correctly.</span>&quot;</p></body></html>"))
        self.radioButton_correction.setText(_translate("DialogLoadTsfFile", "Time series represent corrections for gravity observations"))
        self.checkBox_overwrite_channel.setToolTip(_translate("DialogLoadTsfFile", "Overwrite existing channels with the same location (= station name) when loading a TSF file."))
        self.checkBox_overwrite_channel.setText(_translate("DialogLoadTsfFile", "Overwrite existing channels with identical locations"))



"""
Output functions of gravtools.
This module is part of the gravtools package.
Author: Andreas Hellerschmied
"""

import datetime as dt

# Imports from other gravtools modules:
from bev_legacy import settings


def write_nsb_file(stat_df, instrument_id, path_name_nsb_file):
    """
    Schreiben der Input Datei für die NSDB (<session_name>.nsb).

    :param stat_df: Station dataframe
    :param path_name_nsb_file: path and name of output file
    :param instrument_id: instrument ID of the used gravimeter
    :return:
    """
    fobj_out = open(path_name_nsb_file, 'w')

    # Loop over stations:
    for i, values in stat_df.iterrows():
        out_str = '{:10s} {:8s}  {:9.0f}{:4.0f} {:1s} {:4s}{:5.0f}{:5.0f}\n'.format(
            values.punktnummer,
            values.date.strftime('%Y%m%d'),
            values.g_abs_full_mugal,
            values.sig_g_abs_mugal,
            settings.GRAVIMETER_KZ_BEV[instrument_id],
            instrument_id,
            values.dhb_m * 100,
            values.dhf_m * 100,
        )
        fobj_out.write(out_str)
    fobj_out.close()


def write_schwaus_protcol(obs_df, stat_df, pol_coef, pol_coef_sig_mugal, session_name, path_save_file,
                          instrument_id, skalierung=None, g0_mean_mugal=None, sig_g0_mean_mugal=None):
    """
    Create processing and results protocol for the analyzed gravity observations.

    :param obs_df:
    :param stat_df:
    :param pol_coef:
    :param pol_coef_sig_mugal:
    :param session_name:
    :param path_save_file:
    :param skalierung:
    :return:
    """

    file_path_name_str =  path_save_file + session_name + '_prot.txt'
    td = (obs_df.index.max() - obs_df.index.min())  # Session duration
    num_of_obs = obs_df.shape[0]
    gravimeter_str = settings.GRAVIMETER_ID_BEV[instrument_id]
    grav_height_corr_cm = settings.GRAVIMETER_REFERENCE_HEIGHT_CORRECTIONS_m[gravimeter_str] * 1e2  # in [cm]

    with open(file_path_name_str, 'w') as f:
        f.write('### Protokoll zur Gravimetrie Auswertung ({}) ###\n'.format(dt.datetime.now().strftime('%Y-%m-%d, '
                                                                                                      '%H:%M')))
        f.write('\n')
        f.write(' - Messdatei:          {}\n'.format(session_name))
        f.write('   - Start der Messung:  {} [{}]\n'.format(obs_df.index.min().strftime('%Y-%m-%d, %H:%M'),
                                                            obs_df.index.min().tzname()))
        f.write('   - Ende der Messung:   {} [{}]\n'.format(obs_df.index.max().strftime('%Y-%m-%d, %H:%M'),
                                                            obs_df.index.min().tzname()))
        f.write('   - Messdauer:          {} Tage, {} h, {} min\n'.format(td.days, td.seconds//3600,
                                                                       (td.seconds//60) % 60))
        f.write('   - Anz. Beobachtungen: {} ({:3.1f}/h)\n'.format(num_of_obs, num_of_obs / (td.seconds/3600)))
        f.write('   - Skalierung:         {}\n'.format(skalierung))
        f.write('   - Annahme: Beobachtungen bereits Gezeiten-korrigiert!\n')
        f.write('   - Beobachtungen mittels VG (und dhf, dhb) auf Festpunkt-Niveau reduziert.')
        f.write(' - Gravimeter:         {}\n'.format(settings.GRAVIMETER_ID_BEV[instrument_id]))
        f.write('   - Bezugshöhen (dhf, dhf) korrigiert um: {:+5.1f} cm\n'.format(grav_height_corr_cm))
        f.write(' - Stationen gesamt:   {}\n'.format(stat_df.shape[0]))
        f.write('   - Drift Stationen:    {}\n'.format(stat_df[stat_df['is_drift_point'] == True].shape[0]))
        f.write('   - ÖSGN Stationen:     {}\n'.format(stat_df[stat_df['is_oesgn'] == True].shape[0]))
        f.write('   - Mittlere Breite:    {:6.3f}°\n'.format(stat_df.breite_deg.mean()))
        f.write('   - Mittlere Länge:     {:6.3f}°\n'.format(stat_df.laenge_deg.mean()))
        f.write('\n')
        f.write('\n')
        f.write('##### Drift-Korrektur #####\n')
        f.write(' - Anpassung eines Drift-Polynoms (Grad {}) mittels multipler linearer Regression.\n'.format(
            len(pol_coef)))
        f.write('   - Polynomkoeffizienten:\n')
        for degree, coef in pol_coef.items():
            f.write('     - A{} = {:8.6f} µGal/h ({:8.6f} mGal/Tag)\n'.format(degree, coef, coef*24/1e3))
        # Mittere quadratische Abweichung der Drift-korrigierten Lesungen vom Lesungs-Schätzwert an der Station
        f.write('     - Mittlere quadr. Abweichung = {:8.6f} µGal\n'.format(pol_coef_sig_mugal))

        f.write('\n')
        f.write('### (Reduzierte) Lesungen und Schätzwerte an den Stationen ###\n')
        f.write('Zeit   Lesung   dhf   VG       Red. Lesung   Abw.\n')
        f.write('[{}]  [µGal]   [cm]  [µGal/m] [µGal]        [µGal]\n'.format(obs_df.index.tz.zone))
        f.write('\n')
        for i, stats in stat_df.iterrows():
            stat_name = stats.punktnummer

            for j, obs in obs_df.loc[obs_df['punktnummer'] == stat_name].iterrows():
                f.write('{:2d}:{:2d}  {:7.0f}  {:4.1f}  {:5.1f}    {:9.1f}   {:5.1f}\n'.format(
                    obs.name.hour,
                    obs.name.minute,
                    obs.g_mugal,
                    obs.dhf_m * 100,
                    obs.vg_mugal,
                    obs.g_red_mugal,
                    obs.abw_mugal))
                pass

            f.write(' => {:10s} : {:9.1f} µGal (+/- {:5.1f} µGal)\n'.format(stat_name,
                                                                            stats.g_est_mugal,
                                                                            stats.sig_g_est_mugal))
            if stats.is_oesgn:
                f.write('    Koordinaten ÖSGN: {:8.4f}°, {:8.4f}° \n'.format(stats.breite_deg, stats.laenge_deg))
            f.write('\n')

        f.write('\n')
        f.write('##### Absolute Schwere mit Bezug zum ÖSGN #####\n')
        f.write(' - Mittelwert der Abweichungen an ÖSGN-Stationen:\n')
        f.write('     go_mean = {:9.1f} +/- {:5.1f} µGal\n'.format(g0_mean_mugal, sig_g0_mean_mugal))
        f.write('\n')
        f.write('PunktNr.    g_ÖSGN       g_relativ         g0                 g_abs            Verb.\n'.format(obs_df.index.tz.zone))
        f.write('            [µGal]       [µGal]            [µGal]             [µGal]           [µGal]\n')
        # f.write('\n')
        for i, stats in stat_df.iterrows():
            if stats.is_oesgn:
                f.write('{:10s}  {:6.0f}+-{:3.0f}  {:9.1f}+-{:5.1f}  {:9.1f}+-{:5.1f}  {:8.1f}+-{:5.1f}  ' \
                        '{:5.1f}\n'.format(stats.punktnummer,
                                              stats.g_oesgn_mugal,
                                              stats.g_sig_oesgn_mugal,
                                              stats.g_est_mugal,
                                              stats.sig_g_est_mugal,
                                              stats.g0_mugal,
                                              stats.sig_g0_mugal,
                                              stats.g_abs_mugal,
                                              stats.sig_g_abs_mugal,
                                              stats.verb_mugal))
            else:
                f.write('{:10s}               {:9.1f}+-{:5.1f}                     {:8.1f}+-{:5.1f}\n'.format(
                                           stats.punktnummer,
                                           stats.g_est_mugal,
                                           stats.sig_g_est_mugal,
                                           stats.g_abs_mugal,
                                           stats.sig_g_abs_mugal,))

    pass




"""
Gravity network adjustment described by Reilly (1970).
Output functions of gravtools.
Author: Andreas Hellerschmied
"""

import sys
import numpy as np

# Imports from other gravtools modules:
from bev_legacy import settings
from bev_legacy import init


def adjust_reilly1970(obs_df, stat_df):
    """
    Adjustment of gravity meter observations according to Reilly 1970.
    Estimated parameters:

     - offset and linear drift per instrument
     - correction of the calibration constant of each instrument
     - gravity at observed stations

    :return:
    """

    # datum_stations_list = ['0-059-21']
    datum_stations_list = []
    stat_df['is_datum'] = False
    obs_df['is_datum'] = False
    if not datum_stations_list:  # is not empty
        # flag_test = False
        # if flag_test:
        # All ÖSGN stations are datum stations
        stat_df['is_datum'] = stat_df.is_oesgn
        obs_df.loc[~obs_df.g_oesgn_mugal.isna(), 'is_datum'] = True
    else:
        # Use datum stations from list:
        for stat in datum_stations_list:
            stat_df.loc[stat_df['punktnummer'] == stat, 'is_datum'] = True
            obs_df.loc[obs_df['punktnummer'] == stat, 'is_datum'] = True
            # Check, if all datum stationsade ÖSGN stations:
            if not stat_df[stat_df['punktnummer'] == stat].is_oesgn.bool():
                print('ERROR: the datum station {} is no ÖSGN station!'.format(stat))

    obs_df = obs_df.sort_values('is_datum')  # Sort: [obs of new stations; obs of datum stations]
    obs_df = obs_df.reset_index()  # enumeration as index instead of time-tag
    stat_df = stat_df.sort_values('is_datum')  # Sort: [obs of new stations; obs of datum stations]

    # Lookup-table for station numbers:
    # - e.g. indicating the position of station-related matrix entries
    stat_num_dict = stat_df['punktnummer'].to_dict()
    stat_num_dict = dict((v, k) for k, v in stat_num_dict.items())  # Exchange keys and values

    # Lookup-table for gravimeters:
    gravimeter_type_dict = {}
    gravimeter_type = obs_df.gravimeter_typ.unique()
    for num, grav_type in enumerate(gravimeter_type):
        gravimeter_type_dict[grav_type] = num

    # Get parameters:
    num_gravimeter = obs_df.gravimeter_typ.unique().shape[0]
    num_est_para_per_instr = 2  # number of estimated parameters per gravimeter
    num_est_instr_para = num_est_para_per_instr * num_gravimeter  # total number of estimated instrument parameters
    num_obs = obs_df.shape[0]
    num_stations = stat_df.shape[0]
    num_datum_stations = stat_df[stat_df['is_datum'] == True].shape[0]  # all ÖSGN stations
    num_new_stations = num_stations - num_datum_stations

    # Preallocate matices and vectors:
    P = np.zeros([num_obs, num_stations])
    Q = np.zeros([num_datum_stations, num_stations])
    T = np.zeros([num_obs, num_est_instr_para])
    Z = np.zeros([num_datum_stations, num_est_instr_para])
    y = np.empty(num_obs)
    y = y.reshape(-1, 1)
    y[:] = np.nan

    # Set up vectors:
    h = stat_df[stat_df.is_datum].g_oesgn_mugal.to_numpy()
    h = h.reshape(-1, 1)

    # Set up individual matrices:
    for i_obs, values in obs_df.iterrows():
        if not values['is_datum']:  # Observed station is not a datum station
            # print(values['punktnummer'])
            idx_stat = stat_num_dict[values['punktnummer']]  # Get index of observed station
            P[i_obs, idx_stat] = 1
            y[i_obs] = values['g_red_mugal']
        else:  # Observed station is a datum station
            y[i_obs] = values['g_red_mugal'] - values['g_oesgn_mugal']
        for grav_type, idx_start in gravimeter_type_dict.items():
            T[i_obs, idx_start * num_est_para_per_instr] = 1  # a_k
            T[i_obs, idx_start * num_est_para_per_instr + 1] = values['dt_h']  # b_k
            if num_est_para_per_instr == 3:
                T[i_obs, idx_start * num_est_para_per_instr + 2] = values['g_mugal'] - values['tide_mugal']  # beta_k

    i_col = 0
    for i_stat, values in stat_df.iterrows():
        if values['is_datum']:
            Q[i_col, stat_num_dict[values['punktnummer']]] = 1
            i_col = i_col + 1

    # combine matrics and build equation system:
    # - Formulation by Reilly 1970:
    # N11 = P.transpose().dot(P) + Q.transpose().dot(Q)
    # N12 = -P.transpose().dot(T)
    # N21 = -T.transpose().dot(P)
    # N22 = T.transpose().dot(T)
    # N1 = np.concatenate((N11, N12), axis=1)
    # N2 = np.concatenate((N21, N22), axis=1)
    # N = np.concatenate((N1, N2), axis=0)
    #
    # B11 = P.transpose().dot(y) + Q.transpose().dot(h)
    # B21 = -T.transpose().dot(y)
    # B = np.concatenate((B11, B21), axis=0)
    #
    # # Solve equation system:
    # N_inv = np.linalg.inv(N)
    # x = N_inv.dot(B)

    # - Standard formulation (same results as Reilly 1970)
    A = np.concatenate((np.concatenate((P, -T), axis=1), np.concatenate((Q, Z), axis=1)), axis=0)
    l = np.concatenate((y, h), axis=0)
    N = A.transpose().dot(A)

    print(' - Rank of N: {}'.format(np.linalg.matrix_rank(N)))
    print(' - Rank of A: {}'.format(np.linalg.matrix_rank(A)))
    print(' - Size of N: {} x {}'.format(N.shape[0], N.shape[0]))
    print(' - Rangdefizit: {}'.format(N.shape[0] - np.linalg.matrix_rank(N)))

    N_inv = np.linalg.inv(N)
    x = N_inv.dot(A.transpose().dot(l))

    # Retrive and print results:
    #
    print('Geschätzte Schwere an Stationen:')
    for stat, num in stat_num_dict.items():
        print(' - {}: {:10.3f} µGal'.format(stat, x[num].item()))
    print('Sonstige Parameter:')
    for i in range(num_stations, len(x)):
        print(' - a{}: {:15.3f}'.format(i - num_stations + 1, x[i].item()))


def main(path_oesgn_table, name_oesgn_table, path_obs_file, name_obs_file, out_path):

    obs_df, stat_df, df_oesgn, obs_info_dict = init.read_and_prep_data(path_oesgn_table,
                                                                       name_oesgn_table,
                                                                       path_obs_file,
                                                                       name_obs_file)

    # Adjustment
    adjust_reilly1970(obs_df, stat_df)

    print('...finished!')


# Run as standalone program:
if __name__ == "__main__":

    if len(sys.argv) == 1:
        print('Eingangsparameter aus options.py bezogen (Default-Parameter).')
        path_oesgn_table = settings.PATH_OESGN_TABLE
        name_oesgn_table = settings.NAME_OESGN_TABLE
        path_obs_file = settings.PATH_OBS_FILE_BEV
        name_obs_file = settings.NAME_OBS_FILE_BEV
        out_path = settings.OUT_PATH

    elif len(sys.argv) == 2:  # Name of obervation file as input argument
        name_obs_file = sys.argv[1]
        # Default-parameters from options-file:
        path_oesgn_table = settings.PATH_OESGN_TABLE
        name_oesgn_table = settings.NAME_OESGN_TABLE
        path_obs_file = settings.PATH_OBS_FILE_BEV
        out_path = settings.OUT_PATH

    else:
        print('Error: Invalid number of input arguments!')
        exit()

    # Run init module:
    main(path_oesgn_table,
         name_oesgn_table,
         path_obs_file,
         name_obs_file,
         out_path)

else:
    # not run as standalone program, but as module
    pass